import numpy as np

from pydtnn.backends import PromoteToBackend
from pydtnn.layer_base import LayerBase
from pydtnn.utils import find_component
from pydtnn.utils.constants import Array


class Optimizer[T: Array](PromoteToBackend):
    """
    Optimizer abstract base class
    """

    def __init__(self, learning_rate: float = 1e-2, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__()
        self.learning_rate: float = learning_rate
        self.dtype: np.dtype = dtype
        self.context = dict[int, dict[str, int | T]]()

    def initialize(self, list_layers: list[LayerBase]) -> None:
        raise NotImplementedError("method \"initialize\" of an Optimizer's child class is not implemented")

    def update(self, layer: LayerBase) -> None:
        raise NotImplementedError("method \"update\" of an Optimizer's child class is not implemented")


def select(name: str) -> type[Optimizer]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



import numpy as np

from pydtnn.backends.cpu.layers.abstract.block_layer import AbstractBlockLayerCPU
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.activations.relu import Relu
from pydtnn.layers.feed_forward import FeedForward


class FeedForwardCPU(FeedForward[np.ndarray], AbstractBlockLayerCPU):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.FC_1 = FC(shape=(self.d_ff,))
        self.relu = Relu()
        self.dropout = Dropout(rate=self.dropout_rate)
        self.FC_2 = FC(shape=(self.shape[-1],))
        self.paths = [[self.FC_1, self.relu, self.dropout, self.FC_2]]

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)

        # Initialize all sublayers
        for layer in self.children:
            layer.init_backend_from_model(self.model)

        self.FC_1.initialize(prev_shape=(self.shape[-1],), x=x)
        self.relu.initialize(prev_shape=(self.d_ff,), x=self.FC_1.y)
        self.dropout.initialize(prev_shape=(self.d_ff,), x=self.relu.y)
        self.FC_2.initialize(prev_shape=(self.d_ff,), x=self.dropout.y)

        self.y = self.FC_2.y
        self.dx = self.FC_1.dx

        for layer in self.children:
            self.fwd_time += layer.fwd_time
            self.bwd_time += layer.bwd_time
            self.nparams += layer.nparams

    def initialize_block_layer(self):
        pass

    def forward(self, x):
        x = self.FC_1.forward(x)
        x = self.relu.forward(x)
        x = self.dropout.forward(x)
        x = self.FC_2.forward(x)
        return x

    def backward(self, dy):
        dx = self.FC_2.backward(dy)
        dx = self.dropout.backward(dx)
        dx = self.relu.backward(dx)
        dx = self.FC_1.backward(dx)
        return dx



from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.flatten import Flatten
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum

import numpy as np


class FlattenCPU(Flatten[np.ndarray], LayerCPU):

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        y: np.ndarray = x.reshape((x.shape[0], *self.shape), order="C", copy=None)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return y
    # ---

    def backward(self, dy: np.ndarray) -> np.ndarray:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_RESHAPE_DX)
        dx: np.ndarray = dy.reshape((dy.shape[0], *self.prev_shape), order="C", copy=None)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return dx
    # ---



import numpy as np

from pydtnn.backends.cpu.losses.loss import LossCPU
from pydtnn.losses.kl_divergence import KLDivergence


class KLDivergenceCPU(KLDivergence[np.ndarray], LossCPU):

    def compute(self, y_pred, y_targ, batch_size):
        # loss = np.abs(y_targ * (np.log(np.abs(y_targ / (y_pred + self.eps)) + 1)))
        # loss = np.sum(loss) / y_pred.shape[0]
        # dx = - pred / target # Respecto a Target
        dx = np.log(np.abs(y_targ/(y_pred + self.eps)) + 1)  # Respecto a prediction
        dx = dx / batch_size
        loss = np.sum(dx)
        return loss, dx



from copy import deepcopy

from pydtnn.layers.conv_2d import Conv2D
from pydtnn.model import Model
from pydtnn.tests.abstract.common import D
from pydtnn.tests.abstract.common import Params
from pydtnn.tests.abstract.conv_2d_common import Conv2DCommonTestCase
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.initializers import glorot_uniform, zeros


class Conv2DConvGemmTestCase(Conv2DCommonTestCase):
    """
    Tests that Conv2D with conv_gemm leads to the same results than Conv2d with mm and i2c.T
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global Conv2DCommonTestCase
    del Conv2DCommonTestCase

    @staticmethod
    def _get_layers(d: D, deconv=False, trans=False) -> tuple[Conv2D, Conv2D]:
        params = Params()
        params.tensor_format = TensorFormat.NCHW.upper()
        params.batch_size = d.b
        params.conv_variant = "i2c"
        model_i2c = Model(**vars(params))
        model_i2c.mode = Model.Mode.TRAIN
        params_gc = deepcopy(params)
        params_gc.conv_variant = "gemm"
        model_cg = Model(**vars(params_gc))
        model_cg.mode = Model.Mode.TRAIN
        conv2d_i2c = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                            padding=(d.vpadding, d.hpadding),
                            stride=(d.vstride, d.hstride),
                            dilation=(d.vdilation, d.hdilation),
                            use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        conv2d_i2c.init_backend_from_model(model_i2c)
        conv2d_cg = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                           padding=(d.vpadding, d.hpadding),
                           stride=(d.vstride, d.hstride),
                           dilation=(d.vdilation, d.hdilation),
                           use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        conv2d_cg.init_backend_from_model(model_cg)
        for layer in (conv2d_i2c, conv2d_cg):
            layer.initialize(prev_shape=model_i2c.encode_shape((d.c, d.h, d.w)))
        # Set the same initial weights and biases to both layers
        conv2d_cg.weights = conv2d_i2c.weights.copy()
        conv2d_cg.biases = conv2d_i2c.biases.copy()
        return conv2d_i2c, conv2d_cg



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def div_arrays_set_if_zero[T: _npDT](dividend: _npDT_1Dims[T],
                                     divider: _npDT_1Dims[T],
                                     default_value: T):  # type: ignore
    """
    This function execute a element wise division between dividend and divider ("dividend / divider"), but if divider is 0, then the result is 0 too.
    Example:
        dividend = [-1, 1, 0, 10, -29, 3, 0, 0]
        divider = [0, 40, 0, 0, 3, 0, 10, -30]
        default_value = 0

        result = [0, 1/40, 0, 0, -29/3, 0, 0, 0]

    Another example:
        dividend = [-1, 1, 0, 10, -29, 3, 0, 0]
        divider = [0, 40, 0, 0, 3, 0, 10, -30]
        default_value = 33.3

        result = [33.3, 1/40, 33.3, 33.3, -29/3, 33.3, 0, 0]

    Args:
        dividend (npDT_1Dims): The division's dividend.
        divider (npDT_1Dims): The division's divider.
        default_value (npDT): The va

    Returns:
        Nothing. The value is stores in "dividend".
    """



import numpy as np
import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.elementwise import ElementwiseKernel  #type: ignore

from pydtnn.backends.gpu.optimizers.optimizer import OptimizerGPU
from pydtnn.optimizers.sgd import SGD
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class SGDGPU(SGD[TensorGPU], OptimizerGPU):
    """
    SGDGPU optimizer
    """

    def __init__(self, learning_rate=1e-2, momentum=0.9, nesterov=False, decay=0.0, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__(learning_rate, momentum, nesterov, decay, dtype)

        # --- GPU ---
        parameters_gpu = "{T} *w, {T} * dw, {T} * v, float lr, float decay, float momentum".format(T=DTYPE2CTYPE[dtype])
        ops_gpu = {True: "w[i] -= lr * (decay * w[i] + dw[i] + momentum * v[i])",
                   False: "w[i] -= lr * (decay * w[i] + v[i])"}[nesterov]
        operations_gpu = "v[i] = momentum * v[i] + dw[i]; {nesterov_ops};".format(nesterov_ops=ops_gpu)
        
        self.update_gpu = ElementwiseKernel(parameters_gpu, operations_gpu, "SGD_kernel")
        # ------------
    
        # GPU Direct -
        _name = "SGD_kernel_gpudirect"
        code = """
        __global__ void {name}({T} *w, {T} *dw, {T} *v,
                            float lr, float decay, float momentum, int N) 
        {{
            int i = blockIdx.x * blockDim.x + threadIdx.x;
            if (i < N) 
            {{
                v[i] = momentum * v[i] + dw[i];
                {nesterov_ops};
            }}
            }}
        """.format(
            T = DTYPE2CTYPE[dtype],
            nesterov_ops=({True: "w[i] -= lr * (decay * w[i] + dw[i] + momentum * v[i])",
                            False: "w[i] -= lr * (decay * w[i] + v[i])"}[nesterov]),
            name = _name
        )
        
        self.update_gpudirect = SourceModule(code).get_function(_name)
        # ------------
    
    def initialize(self, list_layers: list[LayerGPU]) -> None:
        for layer in list_layers:
            list_grad_vars = list(layer.grad_vars.keys())

            if len(list_grad_vars) != 0:
                self.context[layer.id] = dict[str, gpuarray.GPUArray]()
                for w_ in list_grad_vars:
                    w = getattr(layer, w_)
                    self.context[layer.id]["velocity_%s" % w_] = gpuarray.zeros_like(w.ary, dtype=w.ary.dtype)

    def update(self, layer: LayerGPU):
        for w_, dw_ in layer.grad_vars.items():
            w, dw = getattr(layer, w_), getattr(layer, dw_)
            velocity = self.context[layer.id]["velocity_%s" % w_]
            w: TensorGPU
            dw: TensorGPU
            velocity: gpuarray.GPUArray

            if self.gpudirect:
                n = self.get_batch_size(w)
                threads, blocks = self.get_threads_and_blocks()

                self.update_gpudirect(w.ary.gpudata, dw.ptr_intp, velocity.gpudata,
                                      np.float32(self.learning_rate), np.float32(self.decay),
                                      np.float32(self.momentum), np.int32(n),
                                      grid=(int(blocks), 1, 1), block=(int(threads), 1, 1),
                                      stream=layer.stream_2)
            else:
                n = np.int32(np.prod(w.shape))
                self.update_gpu(w.ary, dw.ary, velocity, np.float32(self.learning_rate),
                                np.float32(self.decay), np.float32(self.momentum),
                                stream=layer.stream_2)



from pydtnn.layers.layer import Layer
from pydtnn.utils.constants import Array
from pydtnn.utils.constants import ArrayShape


class Dropout[T: Array](Layer[T]):
    def __init__(self, rate=0.5):
        super().__init__()
        self.rate = min(1., max(0., rate))

    def initialize(self, prev_shape: ArrayShape, x: T | None):
        super().initialize(prev_shape, x)
        self.shape = prev_shape



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Normalization layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any

# Functionality imports
from pydtnn.layers.batch_normalization import BatchNormalization
import pydtnn.converters.pytorch2pydtnn.common as cm

# ------------------ #


def BatchNorm2d(args: Dict[str, Any]) -> BatchNormalization:
    # https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm2d.html#torch.nn.BatchNorm2d

    # PyTorch attributes:
    # Not used: num_features, affine, track_running_stats
    PYTORCH_EPS = "eps"  # Float
    PYTORCH_MOMENTUM = "momentum"  # Float

    torch_dict_keys = [PYTORCH_MOMENTUM, PYTORCH_EPS]
    # ---- #

    # PyDTNN attributes:
    # Not used: beta, gamma
    PYDTNN_MOMENTUM = "momentum"
    PYDTNN_EPSILON = "epsilon"

    pydtnn_dict_keys = [PYDTNN_MOMENTUM, PYDTNN_EPSILON]
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    return BatchNormalization(**layer_args)



# ONNX operations:
from .operations.implemented_operations import *

from typing import Callable
from pydtnn.layer_base import LayerBase

CONST_NODE = "node"
CONST_OPSET = "opset_version"
CONST_OUPTUS = "outputs"
CONST_ATTRIBUTES = "attributes"
CONST_INPUTS = "inputs"
CONST_ALL_INPUTS = "all_inputs"
CONST_listS_NODES = "lists_nodes"
CONST_WEIGHTS = "weights"
CONST_PREV_LAYERS = "previous_layers"

# Operations to do:
# DenseNet169 - {'Conv', 'BatchNormalization', 'Unsqueeze', 'Add', 'Mul', 'Relu', 'MaxPool', 'AveragePool', 'GlobalAveragePool', 'Concat'}
# ResNet50 - {'Conv', 'MaxPool', 'Relu', 'Add', 'BatchNormalization', 'GlobalAveragePool', 'Gemm', 'Flatten'}
# VGG19 - {'Dropout', 'Gemm', 'Flatten', 'Relu', 'MaxPool', 'BatchNormalization', 'Conv'}
# Union of the ones before - {'Add', 'AveragePool', 'BatchNormalization', 'Concat', 'Conv', 'Dropout', 'Flatten', 'Gemm', 'GlobalAveragePool', 'MaxPool', 'Mul', 'Relu', 'Unsqueeze'}


def pads_from_onnx_to_pydtnn(pads: list[int]) -> tuple[int, int]:  # -> list[tuple[int, int]]:
    # "pads format should be as follow [x1_begin, x2_begin…x1_end, x2_end,…]" from, for example, https://onnx.ai/onnx/operators/onnx__AveragePool.html
    # Onnx: [x1_begin, x2_begin, ..., x1_end, x2_end, ...] ==> "PyDTNN: [(x1_begin, x1_end), (x2_end, x2_begin), ...]"
    # ==> PyDTNN only admits a int or a (vpadding, hpadding) ==> It's assumed that is the first tuple.

    print(f"pads: {pads}")  # TODO: Borrar
    num_pads = len(pads)//2
    _pads = [(0, 0)] * (num_pads)
    for i in range(num_pads):
        _pads[i] = (int(pads[i]), int(pads[i + num_pads]))
    print(f"_pads: {_pads}")  # TODO: Borrar

    return _pads[0]
# --- END pads_from_onnx_to_pydtnn --- #


def not_implemented(name: str) -> Callable:
    # Normal usage of this: switch_pytorch_pydtnn([not_implemented_layer_name])(args)
    def _not_implemented(args: dict[str, Any]) -> None:
        raise NotImplementedError(f"Layer \"{name}\" not implemented - Args received:\n{args} ")
    return _not_implemented
# --- END not_implemented --- #


def switch_onnx_operation_to_pydtnn(name: str) -> Callable[[dict[str, Any]], LayerBase]:
    match name:
        case "Add": return Add
        case "AveragePool": return AveragePool
        case "BatchNormalization": return BatchNormalization
        case "Concat": return Concat
        case "Conv": return Conv
        case "Dropout": return Dropout
        case "Flatten": return Flatten
        case "Gemm": return Gemm
        case "GlobalAveragePool": return GlobalAveragePool
        case "MaxPool": return MaxPool
        case "Mul": return Mul
        case "Relu": return Relu
        case "Unsqueeze": return Unsqueeze
        # Base case:
        case _: return not_implemented(name)
# --- END switch_onnx_operation_to_pydtnn --- #



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Dropout layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any

# Functionality imports
from pydtnn.layers.dropout import Dropout as _Dropout
import pydtnn.converters.pytorch2pydtnn.common as cm

# ------------------- #


def Dropout(args: Dict[str, Any]) -> _Dropout:
    # https://pytorch.org/docs/stable/generated/torch.nn.Dropout.html#torch.nn.Dropout

    # PyTorch attributes:
    # Not used: inplace: Bool
    PYTORCH_P = "p"
    torch_dict_keys = [PYTORCH_P]
    # ---- #

    # PyDTNN attributes:
    PYDTNN_RATE = "rate"
    pydtnn_dict_keys = [PYDTNN_RATE]
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    return _Dropout(**layer_args)



"""Abstract encryption"""

# TODO: Migrate to pyfhel

import math
import operator
import itertools
import dataclasses
from collections import abc
from dataclasses import dataclass

import numpy as np
from pydtnn.utils.constants import ArrayShape


__all__ = (
    "Context",
)


@dataclass(eq=False, order=False, slots=True, frozen=True)
class Ciphertext[C, P: np.number]:
    """Abstract ciphertext"""
    dtype: np.dtype[P]
    shape: ArrayShape
    _chunks: tuple[C, ...] = dataclasses.field(repr=False)

    def _new(self, /, *args, **kwds):
        """Create new operable ciphertext"""
        return self.__class__(
            dtype=self.dtype,
            *args, **kwds
        )

    def _operable(self, other) -> None:
        """Ensure ciphertext is operable"""
        if not isinstance(other, Ciphertext):
            raise NotImplementedError()

        if other.dtype != self.dtype:
            raise TypeError(f"Different underlying types ({other.dtype} != {self.dtype})")

    def __add__(self, other):
        """Add two ciphertexts"""
        self._operable(other)

        if other.shape != self.shape:
            raise TypeError(f"Different underlying shapes ({other.shape} != {self.shape})")

        chunks = tuple(itertools.starmap(operator.add, zip(self._chunks, other._chunks)))

        return self._new(
            shape=self.shape,
            _chunks=chunks
        )


class Context[C]:
    """Abstract context"""
    _cls: type[Ciphertext]

    def __init__(self, poly_degree: int = 13, global_scale: int = 40, security_level: int = 128) -> None:
        """Inizialize context"""
        self._poly_degree = poly_degree
        self._global_scale = global_scale
        self._security_level = security_level

    @property
    def _slots(self) -> int:
        return 2 ** (self._poly_degree // 2)

    def _new[P: np.number](self, /, dtype: np.dtype[P], *args, **kwds) -> Ciphertext[C, P]:
        """Create new operable ciphertext"""
        return self._cls(
            dtype=dtype,
            *args, **kwds
        )

    def _partition(self, obj: np.ndarray) -> abc.Generator[list]:
        """Transform numpy array into batched lists"""
        if obj.size == 0:
            return
        for part in np.array_split(obj.reshape(-1), range(self._slots, obj.size, self._slots)):
            yield part.tolist()

    def _encrypt_chunk(self, chunk: list) -> C:
        """Encode list to ciphertext"""
        raise NotImplementedError()

    def _decrypt_chunk(self, chunk: C) -> list:
        """Decode cypertext to list"""
        raise NotImplementedError()

    def _encrypt[P: np.number](self, obj: np.ndarray[tuple, np.dtype[P]], /, *args, **kwds) -> Ciphertext[C, P]:
        """Encode numpy array to ciphertext"""
        data = tuple(map(self._encrypt_chunk, self._partition(obj)))
        return self._new(
            dtype=obj.dtype,
            shape=obj.shape,
            _chunks=data,
            *args, **kwds
        )

    def _decrypt[P: np.number](self, obj: Ciphertext[C, P]) -> np.ndarray[tuple, np.dtype[P]]:
        """Decode cypertext to numpy array"""
        data = itertools.chain.from_iterable(map(self._decrypt_chunk, obj._chunks))

        return np.fromiter(
            iter=data,
            dtype=obj.dtype,
            count=math.prod(obj.shape)
        ).reshape(obj.shape)

    def encrypt[P: np.number](self, obj: np.ndarray[tuple, np.dtype[P]]) -> Ciphertext[C, P]:
        """Encode numpy array to ciphertext"""
        return self._encrypt(obj)

    def decrypt[P: np.number](self, obj: Ciphertext[C, P]) -> np.ndarray[tuple, np.dtype[P]]:
        """Decode cypertext to numpy array"""
        return self._decrypt(obj)



import operator
import warnings
import numpy as np
from pydtnn.datasets.dataset import Dataset
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydtnn.model import Model

TENSOR_ASSERT = {
    TensorFormat.NCHW: operator.lt,
    TensorFormat.NHWC: operator.gt
}


class CustomDataset(Dataset):
    """
    Custom Dataset

    In-memory dataset.
    Train and Test must have matching types, shapes and dtypes.
    Input must be in NCHW format, output in N (or more) format.
    X must be in a NDArray with `model.tensor_shape` shape and `model.dtype` dtype.
    Y must be in a NDArray with N (or more) and `model.dtype` dtype.
    """

    def __init__(self, model: "Model", x_train: np.ndarray, y_train: np.ndarray,
                 x_test: np.ndarray | None = None, y_test: np.ndarray | None = None,
                 input_shape: ArrayShape | None = None, output_shape: ArrayShape | None = None,
                 force_test_as_validation=False, debug=False):
        if x_test is None or y_test is None:
            if x_test is None and y_test is None:
                x_test = x_train
                y_test = y_train
            else:
                raise ValueError("Both x_test and y_test must be provided or, alternatively, none of them!")

        if input_shape is None:
            _input_shape: ArrayShape = x_train.shape[1:]

        if output_shape is None:
            _output_shape: ArrayShape = y_train.shape[1:]

        if len(x_train.shape) == 3 and not TENSOR_ASSERT[self.model.tensor_format](x_train.shape[0], x_train.shape[2]):
            warnings.warn(f"Dataset x_train.shape {x_train.shape} may not be in {self.model.tensor_format.upper()} format, following the model format!", RuntimeWarning)

        if len(x_test.shape) == 3 and not TENSOR_ASSERT[self.model.tensor_format](x_test.shape[0], x_test.shape[2]):
            warnings.warn(f"Dataset x_test.shape {x_test.shape} may not be in {self.model.tensor_format.upper()} format, following the model format!", RuntimeWarning)

        self.__x_source: list[np.ndarray] = []
        self.__y_source: list[np.ndarray] = []
        # Sources for the training part
        self.__x_source.append(x_train)
        self.__y_source.append(y_train)
        # Sources for the validation part
        if force_test_as_validation:
            self.__x_source.append(x_test)
            self.__y_source.append(y_test)
        else:
            self.__x_source.append(x_train)
            self.__y_source.append(y_train)
        # Sources for the test part
        self.__x_source.append(x_test)
        self.__y_source.append(y_test)

        super().__init__(model,
                         x_train.shape[0],
                         x_test.shape[0],
                         _input_shape,
                         _output_shape,
                         force_test_as_validation=force_test_as_validation,
                         debug=debug)

    def _init_actual_data(self):
        for part in Dataset.Part:
            local_offset = self._local_offset[part]
            local_nsamples = self._local_nsamples[part]
            local_slice = slice(local_offset, local_offset + local_nsamples)
            self._x[part] = self.__x_source[part][local_slice, ...]
            self._y[part] = self.__y_source[part][local_slice, ...]



from pydtnn.layers.adaptive_average_pool_2d import AdaptiveAveragePool2D
from pydtnn.backends.gpu.layers.layer import LayerGPU

# Import from AbstractPool2DLayerGPU
from pydtnn.utils.constants import DTYPE2CTYPE
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.performance_models import im2col_time, col2im_time
import pycuda.gpuarray as gpuarray   # type: ignore
from pycuda.compiler import SourceModule   # type: ignore
from pycuda.driver import Function   # type: ignore

import numpy as np
from pydtnn.utils.tensor import TensorFormat

# --- CONSTANTS --- #
_MACRO_INDEX_FIRST_ELEMENT = "INDEX_FIRST_ELEMENT"
_MACRO_INDEX_LAST_ELEMENT = "INDEX_LAST_ELEMENT"
_MACRO_INDEX_N = "INDEX_N"
_MACRO_INDEX_C = "INDEX_C"
_MACRO_INDEX_H = "INDEX_H"
_MACRO_INDEX_W = "INDEX_W"
_MACRO_SHIFT_POINTER = "SHIFT_POINTER"
_FULL_MACRO_SHIFT_POINTER = f"#define {_MACRO_SHIFT_POINTER}(p, c, h, w, ni, ci, hi, wi)"
_SHIFT_POINTER_NCHW = "p + ((ni * c + ci) * h + hi) * w + wi"
_SHIFT_POINTER_NHWC = "p + ((ni * h + hi) * w + wi) * c + ci"

_FULL_MACRO_INDEX_C_NCHW = f"#define {_MACRO_INDEX_C}(idx, c, h, w) (idx / (h * w)) % c"
_FULL_MACRO_INDEX_H_NCHW = f"#define {_MACRO_INDEX_H}(idx, h, w) (idx / w) % h"
_FULL_MACRO_INDEX_W_NCHW = f"#define {_MACRO_INDEX_W}(idx, w) idx % w"
_DIMENSION_INDEX_CODE_NCHW = \
"""
ci = {macro_index_c}(idx, c, new_h, new_w);
hi = {macro_index_h}(idx, new_h, new_w);
wi = {macro_index_w}(idx, new_w);
"""
_FULL_MACRO_INDEX_H_NHWC = f"#define {_MACRO_INDEX_H}(idx, h, w, c) (idx / (w * c)) % h"
_FULL_MACRO_INDEX_W_NHWC = f"#define {_MACRO_INDEX_W}(idx, w, c) (idx / c) % w"
_FULL_MACRO_INDEX_C_NHWC = f"#define {_MACRO_INDEX_C}(idx, c) idx % c"
_DIMENSION_INDEX_CODE_NHWC = \
"""
hi = {macro_index_h}(idx, new_h, new_w, c);
wi = {macro_index_w}(idx, new_w, c);
ci = {macro_index_c}(idx, c);
"""


class AdaptiveAveragePool2DGPU(AdaptiveAveragePool2D[TensorGPU], LayerGPU):

    def initialize(self, prev_shape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.cuda_fwd_func = self.cuda_adaptive_average_pooling_fwd(dtype=self.model.dtype)
        self.cuda_bwd_func = self.cuda_adaptive_average_pooling_bwd(dtype=self.model.dtype)

        self.initialize_pool_2d_gpu(prev_shape, x)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # NOTE: Will be initalized later.
        self.y = None  # type: ignore

    def cuda_adaptive_average_pooling_fwd(self, dtype: np.dtype) -> Function:

        _func_name = ["cuda_adaptive_average_pooling_fwd"]
        _t = DTYPE2CTYPE[dtype]  # variable Type
        _full_macro_index_c = ""
        _full_macro_index_h = ""
        _full_macro_index_w = ""
        _dimension_index_code = ""
        _full_macro_shift_pointer = [_FULL_MACRO_SHIFT_POINTER]

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                _func_name.append("_nchw")
                _full_macro_shift_pointer.append(_SHIFT_POINTER_NCHW)
                _full_macro_index_c = _FULL_MACRO_INDEX_C_NCHW
                _full_macro_index_h = _FULL_MACRO_INDEX_H_NCHW
                _full_macro_index_w = _FULL_MACRO_INDEX_W_NCHW
                _dimension_index_code = _DIMENSION_INDEX_CODE_NCHW
                # -- END cuda_adaptive_average_pooling_fwd_nchw --
            case TensorFormat.NHWC:
                # NOTE: It has been tested and it return values that seems to make sense,
                #   but they hadn't been compared with other model's output due the format (Torch is NCHW).
                _func_name.append("_nhwc")
                _full_macro_shift_pointer.append(_SHIFT_POINTER_NHWC)
                _full_macro_index_h = _FULL_MACRO_INDEX_H_NHWC
                _full_macro_index_w = _FULL_MACRO_INDEX_W_NHWC
                _full_macro_index_c = _FULL_MACRO_INDEX_C_NHWC
                _dimension_index_code = _DIMENSION_INDEX_CODE_NHWC
                # -- END cuda_adaptive_average_pooling_fwd_nhwc --
            case _:
                raise NotImplementedError(f"{self.model.tensor_format} is not an implemented format.")

        _func_name = "".join(_func_name)
        _full_macro_shift_pointer = "".join(_full_macro_shift_pointer)
        _dimension_index_code = _dimension_index_code.format(macro_index_c=_MACRO_INDEX_C,
                                                             macro_index_h=_MACRO_INDEX_H,
                                                             macro_index_w=_MACRO_INDEX_W)

        code = """
#define {macro_index_n}(idx, N, n) idx * n / N
{full_macro_index_c}
{full_macro_index_h}
{full_macro_index_w}

#define {macro_index_first_element}(index, dim_in, dim_out) (int) ((index * dim_in) / dim_out)
#define {macro_index_last_element}(index, dim_in, dim_out) (int) ((((index + 1) * dim_in) + dim_out - 1) / dim_out)
{full_macro_shift_pointer}

#define TRUE  1
#define FALSE 0

__global__ void {func_name}({T}* x, {T}* y,
                            int n, int c, int h, int w,
                            int new_h, int new_w, int N,
                            int num_active_workers,
                            int num_ops_per_worker,
                            int num_ops_last_worker)
{{
    int idx, ops_remaining;
    int ni, ci, wi, hi, i, j;
    int h_start, h_end, w_start, w_end, elements_h, elements;
    unsigned short first_iteration;
    {T} add;

    idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx >= N) return;
    ops_remaining = ((idx + 1) == num_active_workers) ? num_ops_last_worker : num_ops_per_worker;
    idx *= num_ops_per_worker;

    ni = {macro_index_n}(idx, N, n);
    {dimension_index_code}
    first_iteration = TRUE;

    for(ni = ni;
        (ni < n) && (ops_remaining > 0);
        ni++)
    {{
        for(ci = (first_iteration ? ci : 0);
            (ci < c) && (ops_remaining > 0);
            ci++)
        {{
            for(hi = (first_iteration ? hi : 0);
                (hi < new_h) && (ops_remaining > 0);
                hi++)
            {{
                h_start = {macro_index_first_element}(hi, h, new_h);
                h_end = {macro_index_last_element}(hi, h, new_h);
                elements_h = h_end - h_start;

                for(wi = (first_iteration ? wi : 0), first_iteration = FALSE;
                    (wi < new_w) && (ops_remaining > 0);
                    wi++, ops_remaining--)
                {{
                    w_start = {macro_index_first_element}(wi, w, new_w);
                    w_end = {macro_index_last_element}(wi, w, new_w);
                    elements = elements_h * (w_end - w_start);

                    for(i = h_start, add = ({T}) 0.0; i < h_end; i++)
                        for(j = w_start; j < w_end; j++)
                            add += ({T}) (*({macro_desp_pointer}(x, c, h, w, ni, ci, i, j)) );

                    (*({macro_desp_pointer}(y, c, new_h, new_w, ni, ci, hi, wi))) = ({T}) (add / elements);
                }}
            }}
        }}
    }}
}}
"""

        code = code.format(full_macro_index_c=_full_macro_index_c,
                           full_macro_index_h=_full_macro_index_h,
                           full_macro_index_w=_full_macro_index_w,
                           full_macro_shift_pointer=_full_macro_shift_pointer,
                           macro_index_n=_MACRO_INDEX_N,
                           macro_index_c=_MACRO_INDEX_C,
                           macro_index_h=_MACRO_INDEX_H,
                           macro_index_w=_MACRO_INDEX_W,
                           macro_index_first_element=_MACRO_INDEX_FIRST_ELEMENT,
                           macro_index_last_element=_MACRO_INDEX_LAST_ELEMENT,
                           macro_desp_pointer=_MACRO_SHIFT_POINTER,
                           dimension_index_code=_dimension_index_code,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module

    def cuda_adaptive_average_pooling_bwd(self, dtype: np.dtype) -> Function:

        _func_name = ["cuda_adaptive_average_pooling_bwd"]
        _t = DTYPE2CTYPE[dtype]  # variable Type
        _full_macro_index_c = ""
        _full_macro_index_h = ""
        _full_macro_index_w = ""
        _dimension_index_code = ""
        _full_macro_shift_pointer = [_FULL_MACRO_SHIFT_POINTER]

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                _func_name.append("_nchw")
                _full_macro_shift_pointer.append(_SHIFT_POINTER_NCHW)
                _full_macro_index_c = _FULL_MACRO_INDEX_C_NCHW
                _full_macro_index_h = _FULL_MACRO_INDEX_H_NCHW
                _full_macro_index_w = _FULL_MACRO_INDEX_W_NCHW
                _dimension_index_code = _DIMENSION_INDEX_CODE_NCHW
            case TensorFormat.NHWC:
                _func_name.append("_nhwc")
                _full_macro_shift_pointer.append(_SHIFT_POINTER_NHWC)
                _full_macro_index_h = _FULL_MACRO_INDEX_H_NHWC
                _full_macro_index_w = _FULL_MACRO_INDEX_W_NHWC
                _full_macro_index_c = _FULL_MACRO_INDEX_C_NHWC
                _dimension_index_code = _DIMENSION_INDEX_CODE_NHWC
            case _:
                raise NotImplementedError(f"{self.model.tensor_format} is not an implemented format.")

        _func_name = "".join(_func_name)
        _full_macro_shift_pointer = "".join(_full_macro_shift_pointer)
        _dimension_index_code = _dimension_index_code.format(macro_index_c=_MACRO_INDEX_C,
                                                             macro_index_h=_MACRO_INDEX_H,
                                                             macro_index_w=_MACRO_INDEX_W)

        code = """
#define {macro_index_n}(idx, N, n) idx * n / N
{full_macro_index_c}
{full_macro_index_h}
{full_macro_index_w}

#define {macro_index_first_element}(index, dim_in, dim_out) (int) ((index * dim_in) / dim_out)
#define {macro_index_last_element}(index, dim_in, dim_out) (int) ((((index + 1) * dim_in) + dim_out - 1) / dim_out)
{full_macro_shift_pointer}

#define TRUE  1
#define FALSE 0

__global__ void {func_name}({T}* dx, {T}* dy,
                            int n, int c, int h, int w,
                            int new_h, int new_w, int N,
                            int num_active_workers,
                            int num_ops_per_worker,
                            int num_ops_last_worker)
{{
    int idx, ops_remaining;
    int ni, ci, wi, hi, i, j;
    int h_start, h_end, w_start, w_end, elements_h, elements;
    unsigned short first_iteration;
    {T} delta;

    idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx >= N) return;
    ops_remaining = ((idx + 1) == num_active_workers) ? num_ops_last_worker : num_ops_per_worker;
    idx *= num_ops_per_worker;

    ni = {macro_index_n}(idx, N, n);
    {dimension_index_code}
    first_iteration = TRUE;

    for(ni = ni;
        (ni < n) && (ops_remaining > 0);
        ni++)
    {{
        for(ci = (first_iteration ? ci : 0);
            (ci < c) && (ops_remaining > 0);
            ci++)
        {{
            for(hi = (first_iteration ? hi : 0);
                (hi < h) && (ops_remaining > 0);
                hi++)
            {{
                h_start = {macro_index_first_element}(hi, new_h, h);
                h_end = {macro_index_last_element}(hi, new_h, h);
                elements_h = h_end - h_start;

                for(wi = (first_iteration ? wi : 0), first_iteration = FALSE;
                    (wi < w) && (ops_remaining > 0);
                    wi++, ops_remaining--)
                {{
                    w_start = {macro_index_first_element}(wi, new_w, w);
                    w_end = {macro_index_last_element}(wi, new_w, w);
                    elements = elements_h * (w_end - w_start);

                    delta = ({T}) (*({macro_desp_pointer}(dy, c, new_h, new_w, ni, ci, hi, wi)) / elements);
                    for(i = h_start; i < h_end; i++)
                        for(j = w_start; j < w_end; j++)
                            (*({macro_desp_pointer}(dx, c, h, w, ni, ci, i, j))) += delta;
                }}
            }}
        }}
    }}
}}
"""

        code = code.format(full_macro_index_c=_full_macro_index_c,
                           full_macro_index_h=_full_macro_index_h,
                           full_macro_index_w=_full_macro_index_w,
                           full_macro_shift_pointer=_full_macro_shift_pointer,
                           macro_index_n=_MACRO_INDEX_N,
                           macro_index_c=_MACRO_INDEX_C,
                           macro_index_h=_MACRO_INDEX_H,
                           macro_index_w=_MACRO_INDEX_W,
                           macro_index_first_element=_MACRO_INDEX_FIRST_ELEMENT,
                           macro_index_last_element=_MACRO_INDEX_LAST_ELEMENT,
                           macro_desp_pointer=_MACRO_SHIFT_POINTER,
                           dimension_index_code=_dimension_index_code,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module

    def initialize_pool_2d_gpu(self, prev_shape, x):
        self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)
        self.shape = self.model.encode_shape((self.co, self.ho, self.wo))
        pooling_shape = self.model.encode_shape((self.co, self.ho, self.wo))
        y = gpuarray.empty((self.model.batch_size, *pooling_shape), self.model.dtype)
        self.y = TensorGPU(y, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty(self.x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.fwd_time = \
            im2col_time(m=self.co, n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time = \
            col2im_time(m=self.co, n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)

        if self.pooling_not_needed:
            self.y = x
        else:
            n, c, h, w = self.model.decode_shape(x.shape)

            # NOTE: "num_elements" (or simply "N") is the number of elements to process. Usually it would be np.prod(x.shape),
            #   but in this case we are putting elements in the output instead of processing the input's elements.
            num_elements = np.prod((n, c, self.ho, self.wo), dtype=np.int32)

            total_num_threads = np.prod(self.grid, dtype=np.int32) * np.prod(self.block, dtype=np.int32)

            # If num_elements < total_num_threads, only will work "num_elements" threads. In the other cases will work "total_num_threads" threads.
            num_active_workers = np.int32(min(total_num_threads, num_elements))
            num_ops_per_worker = np.int32((num_elements + num_active_workers - 1) / num_active_workers)
            num_ops_last_worker = np.int32(num_elements - (num_active_workers - 1) * num_ops_per_worker)

            # NOTE: Instead of a number, PyCuda's driver expects "numpy.number"
            self.cuda_fwd_func(x.ary, self.y.ary,
                               np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                               np.int32(self.ho), np.int32(self.wo), num_elements,
                               num_active_workers, num_ops_per_worker, num_ops_last_worker,
                               grid=self.grid, block=self.block,
                               stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        n, c, h, w = self.model.decode_shape(dy.shape)

        num_elements = np.prod((n, c, self.ho, self.wo), dtype=np.int32)

        total_num_threads = np.prod(self.grid, dtype=np.int32) * np.prod(self.block, dtype=np.int32)

        num_active_workers = np.int32(min(total_num_threads, num_elements))
        num_ops_per_worker = np.int32((num_elements + num_active_workers - 1) / num_active_workers)
        num_ops_last_worker = np.int32(num_elements - (num_active_workers - 1) * num_ops_per_worker)

        self.cuda_bwd_func(self.dx.ary, self.y.ary,
                           np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                           np.int32(self.ho), np.int32(self.wo), num_elements,
                           num_active_workers, num_ops_per_worker, num_ops_last_worker,
                           grid=self.grid, block=self.block,
                           stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



import inspect

import numpy as np

from pydtnn.layers.conv_2d import Conv2D
from pydtnn.tests.abstract.common import verbose_test, D
from pydtnn.tests.abstract.common import TestCase
from pydtnn.utils import print_with_header, random
from pydtnn.utils.tensor import TensorFormat, format_transpose


class Conv2DCommonTestCase[T: Conv2D](TestCase):
    """
    Tests that A layer leads to the same results than B layer
    """

    @staticmethod
    def _get_layers(d: D) -> tuple[T, T]:
        raise NotImplementedError()

    @staticmethod
    def _set_state(layer: Conv2D, weights) -> None:
        layer.weights = weights.copy()

    x_2x4 = np.array([[[[1, 2, 4, 8],
                        [16, 32, 64, 128]]]]).astype(np.float32, order='C')

    x_4x4 = np.array([[[[1, 2, 4, 8],
                        [16, 32, 64, 128],
                        [1, 2, 4, 8],
                        [16, 32, 64, 128]]]]).astype(np.float32, order='C')

    x_4x8 = np.array([[[[1, 2, 4, 8, 9, 10, 11, 12],
                        [16, 32, 64, 128, 129, 130, 131, 132],
                        [1, 2, 4, 8, 9, 10, 11, 12],
                        [16, 32, 64, 128, 129, 130, 131, 132]]]]).astype(np.float32, order='C')

    x_8x8 = np.array([[[[11, 12, 13, 14, 15, 16, 17, 18],
                        [21, 22, 23, 24, 25, 26, 27, 28],
                        [31, 32, 33, 34, 35, 36, 37, 38],
                        [41, 42, 43, 44, 45, 46, 47, 48],
                        [51, 52, 53, 54, 55, 56, 57, 58],
                        [61, 62, 63, 64, 65, 66, 67, 68],
                        [71, 72, 73, 74, 75, 76, 77, 78],
                        [81, 82, 83, 84, 85, 86, 87, 88]]]]).astype(np.float32, order='C')

    w_1x1 = np.array([[[[1]]]]).astype(np.float32, order='C')

    w_1x2 = np.array([[[[1, 1]]]]).astype(np.float32, order='C')

    w_2x2 = np.array([[[[1, 1],
                        [1, 1]]]]).astype(np.float32, order='C')

    w_3x3 = np.array([[[[1, 1, 1],
                        [1, 1, 1],
                        [1, 1, 1]]]]).astype(np.float32, order='C')

    def _test_forward_backward(self, d: D, x: np.ndarray, weights: np.ndarray, print_times=False):
        from timeit import timeit
        conv2d_ref, conv2d_test = self._get_layers(d)
        x = conv2d_ref.model.encode_tensor(x).copy()
        if conv2d_ref.model.tensor_format is TensorFormat.NHWC:
            weights = format_transpose(weights, "OIHW", "IHWO").copy()
        self._set_state(conv2d_ref, weights)
        self._set_state(conv2d_test, weights)
        # Forward pass
        y_ref = conv2d_ref.forward(x)
        y_test = conv2d_test.forward(x)
        dy = random.random((d.b, d.kn, d.ho, d.wo)).astype(np.float32, order='C')
        if conv2d_ref.model.tensor_format is TensorFormat.NHWC:
            dy = format_transpose(dy, "NCHW", "NHWC").copy()
        # Backward pass
        dx_ref = conv2d_ref.backward(dy)
        dx_test = conv2d_test.backward(dy)
        # All close?
        dw_allclose = np.allclose(conv2d_ref.dw, conv2d_test.dw)
        dx_allclose = np.allclose(dx_ref, dx_test)
        if verbose_test():
            print_with_header(inspect.stack()[1][3])
            # np.set_printoptions(threshold=50)  # default is 1000
            print(d)
            print("---=[ Forward results ]=---")
            print("y_ref:\n", y_ref)
            print("y_test:\n", y_test)
            print()
            print("---=[ Backward results ]=---")
            print("dx_ref:\n", dx_ref)
            print("dx_test:\n", dx_test)
            print("dw:\n", conv2d_test.dw)
            print("dx allclose: ", dx_allclose)
            print()
            print("---[ dw comparison ]---")
            print("dw_ref.shape:", conv2d_ref.dw.shape)
            print("dw_test.shape: ", conv2d_test.dw.shape)
            print("dw allclose: ", dw_allclose)
            if print_times:
                forward_ref_t = timeit(lambda: conv2d_ref.forward(x), number=10) / 10
                forward_test_t = timeit(lambda: conv2d_test.forward(x), number=10) / 10
                backward_ref_t = timeit(lambda: conv2d_ref.backward(dy), number=10) / 10
                backward_test_t = timeit(lambda: conv2d_test.backward(dy), number=10) / 10
                print()
                print("---[ times comparison ]---")
                print("            ref     test")
                print("         +-------+--------+")
                print(" forward | {:.3f} | {:.3f} |".format(forward_ref_t, forward_test_t))
                print("         +-------+--------+")
                print("backward | {:.3f} | {:.3f} |".format(backward_ref_t, backward_test_t))
                print("         +-------+--------+")
                print("           {:.3f}   {:.3f}  ".format(forward_ref_t + backward_ref_t,
                                                            forward_test_t + backward_test_t))
        # self.assertTrue(np.allclose(y_ref, y_test, rtol=1e-5, atol=1e-6), f"y matrices differ")
        self.assertTrue(np.allclose(y_ref, y_test), "y matrices differ")
        self.assertTrue(dw_allclose, "dw matrices differ")
        self.assertTrue(dx_allclose, "dx return matrices differ")

    def test_forward_defaults(self):
        """
        Test that the default parameters lead to the same solution on the forward step
        """
        d = D()
        conv2d_ref, conv2d_test = self._get_layers(d)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        x = conv2d_ref.model.encode_tensor(x).copy()
        y_ref = conv2d_ref.forward(x)
        y_test = conv2d_test.forward(x)
        if verbose_test():
            print_with_header("test forward defaults")
            print(y_ref)
            print(y_test)
            print("y_ref.shape:", y_ref.shape)
            print("y_test.shape: ", y_test.shape)
        allclose = np.allclose(y_ref, y_test, rtol=1e-5, atol=1e-6)
        self.assertTrue(allclose)

    def test_forward_backward_defaults(self):
        """
        Test that the default parameters lead to the same solution on the backward step
        """
        d = D()
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_handmade_array(self):
        """Tests that manual matrices lead to the same solution"""
        x = self.x_2x4
        weights = self.w_2x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (2, 4)
        d.kh, d.kw = (2, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 1
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_handmade_array_stride2(self):
        """Tests that manual matrices with stride 2 lead to the same solution"""
        x = self.x_2x4
        weights = self.w_2x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (2, 4)
        d.kh, d.kw = (2, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 2
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_larger_handmade_array_stride2(self):
        """Tests that larger manual matrices with stride 2 lead to the same solution"""
        x = self.x_4x4
        weights = self.w_2x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (4, 4)
        d.kh, d.kw = (2, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 2
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_larger_handmade_array_stride3(self):
        """Tests that larger manual matrices with stride 3 lead to the same solution"""
        x = self.x_4x4
        weights = self.w_2x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (4, 4)
        d.kh, d.kw = (2, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 3
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_even_larger_handmade_array_stride3(self):
        """Tests that even larger manual matrices with stride 3 lead to the same solution on i2c and on conv_gemm"""
        x = self.x_4x8
        weights = self.w_2x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (4, 8)
        d.kh, d.kw = (2, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 3
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_even_larger_handmade_array_stride3_filter1x2(self):
        """Tests that even larger manual matrices with stride 3 lead to the same solution on i2c and on conv_gemm"""
        x = self.x_4x8
        weights = self.w_1x2
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (4, 8)
        d.kh, d.kw = (1, 2)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 3
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_even_larger_handmade_array_stride3_filter1x1(self):
        """Tests that even larger manual matrices with stride 3 lead to the same solution on i2c and on conv_gemm"""
        x = self.x_4x8
        weights = self.w_1x1
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (4, 8)
        d.kh, d.kw = (1, 1)
        d.vpadding = d.hpadding = 0
        d.vstride = d.hstride = 3
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_even_larger_handmade_array_stride12(self):
        """Tests that even larger manual matrices with strides 1, 2 lead to the same solution on i2c and on conv_gemm"""
        x = self.x_8x8
        weights = self.w_3x3
        d = D()
        d.kn = d.b = d.c = 1
        d.h, d.w = (8, 8)
        d.kh, d.kw = (3, 3)
        d.vpadding = d.hpadding = 0
        d.vstride = 1
        d.hstride = 2
        d.vdilation = d.hdilation = 1
        self._test_forward_backward(d, x, weights)

    def test_forward_backward_alexnet_cifar10_first_conv2d(self):
        """Tests that the AlexNet cifar10 first Conv2d lead to the same solution on i2c and on conv_gemm"""
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (64, 3, 3)
        d.c, d.h, d.w = (3, 32, 32)
        d.vpadding, d.hpadding = (1, 1)
        d.vstride, d.hstride = (2, 2)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)

    def test_forward_backward_alexnet_imagenet_first_conv2d(self):
        """Tests that the AlexNet ImageNet first Conv2d lead to the same solution on i2c and on conv_gemm"""
        # id;height;width;channels;kernel_height;kernel_width;kernel_num;stride;padding
        # 2;227;227;3;11;11;96;4;0
        d = D()
        d.b = 64
        d.kn, d.kh, d.kw = (96, 11, 11)
        d.c, d.h, d.w = (3, 227, 227)
        d.vpadding, d.hpadding = (1, 1)
        d.vstride, d.hstride = (4, 4)
        d.vdilation, d.hdilation = (1, 1)
        x = random.random((d.b, d.c, d.h, d.w)).astype(np.float32, order='C')
        weights = random.random((d.kn, d.c, d.kh, d.kw)).astype(np.float32, order='C')
        self._test_forward_backward(d, x, weights, print_times=True)



"""
Performance tests for transposing matrices

For running the tests run:
    python best_conv2d_transpose_0231.py
"""

from pydtnn.profilers.best_of_profiler import BestOfProfiler
from pydtnn.tests.abstract.common import alexnet_layers
from pydtnn.utils.best_transpose_1023 import best_transpose_1023
from pydtnn.utils import random


def main():
    layers = alexnet_layers
    bop = BestOfProfiler("Transpose 1023 comparison", best_transpose_1023)
    for layer in layers:
        d0, d1, d2, d3 = layer.shape
        original = random.random((d0, d1, d2, d3)).astype(layer.dtype, order="C", copy=None)
        bop(original)
    bop.print_results()


if __name__ == "__main__":
    main()



from collections import abc

from pydtnn.activations.activation import Activation
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
from pydtnn.utils.constants import ArrayShape
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU

try:
    from pydtnn.libs.libmpi import MPI
except Exception as e:
    pass

try:
    import pydtnn.libs.libnccl as nccl
except Exception as e:
    pass


class ActivationGPU(Activation[TensorGPU]):
    """
    Extends an Activation class with the attributes and methods required by GPU Activations.

    The next methods are copied from LayerGPU:
      * reduce_weights_async()
      * wait_allreduce_async()
      * reduce_weights_sync()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The following attributes will be initalized later.
        self.x: TensorGPU = None  # type: ignore
        self.dx: TensorGPU = None  # type: ignore
        self.grid = None
        self.block = None

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU):
        super().initialize(prev_shape, x)
        self.grid = self.model.cuda_grid
        self.block = self.model.cuda_block

    def reduce_weights_async(self, gradient=True):
        if not self.model.comm:
            return
        self.reqs_allred = {}

        # if self.model.enable_cudnn:
        #     if self.model.enable_nccl or self.model.gpudirect:
        #        self.model.stream.synchronize()
        #     else:
        #        self.stream_2.synchronize()

        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            dw = getattr(self, dw_)

            if self.model.enable_nccl:
                self.model.stream.synchronize()
                dw *= self.model.rank_weight
                # TODO: crypt
                nccl.ncclAllReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                                   nccl.RedOp.Sum, comm=self.model.nccl_comm,
                                   stream=self.stream_2.handle)

                # # Hierarchical mode NCCL + MPI
                # if len(self.model.inter_ranks) == 1:
                #     nccl.ncclAllReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                        nccl.RedOp.Sum, comm=self.model.nccl_comm,
                #                        stream=self.stream_2.handle)
                #
                # else:
                #     # Hierarchical allreduce - Phase 1: ncclReduce + Iallreduce
                #     nccl.ncclReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                     nccl.RedOp.Sum, root=0, comm=self.model.nccl_comm,
                #                     stream=self.stream_2.handle)
                #
                #     if self.model.rank in self.model.inter_ranks:
                #         if not self.model.gpudirect:
                #             dw.ary.get_async(self.stream_2, dw_cpu)
                #
                #         self.stream_2.synchronize()
                #         req = self.model.inter_comm.Iallreduce(MPI.IN_PLACE, dw_cpu, op=MPI.SUM)

            else:  # Without NCCL

                # We have asynchronously moved the dw and db to dw_cpu and db_cpu in stream_2
                # so we need to synchronize stream_2 before performing Allreduce.
                # In GPU direct we have to synchronize the main stream to ensure dw and db are ready.

                if not self.model.gpudirect:
                    self.stream_2.synchronize()
                else:
                    self.model.stream.synchronize()

                dw_cpu = getattr(self, f"{dw_}_cpu")
                dw_cpu *= self.model.rank_weight
                if self.model.crypt:
                    dw_cpu = self.model.crypt.encrypt(dw_cpu)
                if isinstance(dw_cpu, abc.Buffer):
                    req = self.model.comm.Iallreduce(MPI.IN_PLACE, dw_cpu, op=MPI.SUM)
                else:
                    req = self.model.comm.iallreduce(dw_cpu, op=MPI.SUM)
                self.reqs_allred[dw_] = req

    def wait_allreduce_async(self, gradient=True):
        if not self.model.comm:
            return

        for w_, dw_ in self.grad_vars.items():
            if self.model.enable_nccl:
                self.model.stream.synchronize()
                dw = getattr(self, dw_)
                # TODO: decrypt
                setattr(self, dw_, dw)
            else:
                dw_ = dw_ if gradient else w_
                res = self.reqs_allred[dw_].wait()
                if res is None:
                    dw = getattr(self, dw_)
                else:
                    dw = res
                if self.model.crypt:
                    dw = self.model.crypt.decrypt(dw)  #type: ignore
                setattr(self, dw_, dw)

                # # Hierarchical mode NCCL + MPI
                # if self.model.enable_nccl:
                #     if len(self.model.inter_ranks) == 1:
                #         # Do nothing, Allreduce was already completed in phase 1
                #         pass
                #     else:
                #         # Hierarchical allreduce - Phase 2: wait + ncclBroadcast
                #         if self.model.rank in self.model.inter_ranks:
                #             self.reqs_allred[dw_].wait()
                #             if not self.model.gpudirect:
                #                 dw.ary.set_async(dw_cpu, self.stream_2)
                #
                #         nccl.ncclBroadcast(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                            root=0, comm=self.model.nccl_comm,
                #                            stream=self.stream_2.handle)

                if not self.model.gpudirect:
                    dw: TensorGPU = getattr(self, dw_)
                    dw_cpu = getattr(self, f"{dw_}_cpu")

                    # If there is no CUDA-aware MPI, copy data back to GPU
                    dw.ary.set_async(dw_cpu, self.stream_2)

    def reduce_weights_sync(self, gradient=True):
        if not self.model.comm:
            return

        for w_, dw_ in self.grad_vars.items():
            dw_ = dw_ if gradient else w_
            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT],
                                          [self.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.ALLREDUCE_DW,
                                           self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.OPS_ALLREDUCE_DW])
            # stream = self.stream_2.handle)
            dw: TensorGPU = getattr(self, dw_)

            if self.model.enable_nccl:
                dw *= self.model.rank_weight # TODO: Check this!!
                self.stream_2.synchronize()
                # TODO: crypt
                nccl.ncclAllReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                                   nccl.RedOp.Sum, comm=self.model.nccl_comm,
                                   stream=self.stream_2.handle)
                self.stream_2.synchronize()
                # TODO: decrypt

                # # Hierarchical mode NCCL + MPI
                # if len(self.model.inter_ranks) == 1:
                #     # Only one node involved, perform ncclAllreduce across intra-node GPUs
                #     nccl.ncclAllReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                        nccl.RedOp.Sum, comm=self.model.nccl_comm,
                #                        stream=self.stream_2.handle)
                # else:
                #     # Hierarchical allreduce: ncclReduce + Allreduce + ncclBroadcast
                #     nccl.ncclReduce(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                     nccl.RedOp.Sum, root=0, comm=self.model.nccl_comm,
                #                     stream=self.stream_2.handle)
                #
                #     self.stream_2.synchronize()
                #     if self.model.rank in self.model.inter_ranks:
                #         if self.model.gpudirect:
                #             self.model.inter_comm.Allreduce(MPI.IN_PLACE, dw_cpu, op=MPI.SUM)
                #         else:
                #             dw_cpu = dw.ary.get()
                #             self.model.inter_comm.Allreduce(MPI.IN_PLACE, dw_cpu, op=MPI.SUM)
                #             dw.ary.set_async(dw_cpu, self.stream_2)
                #
                #     nccl.ncclBroadcast(dw.ptr, dw.ptr, dw.size, self.model.nccl_type,
                #                        root=0, comm=self.model.nccl_comm,
                #                        stream=self.stream_2.handle)

            else:  # Without NCCL

                # We have asynchronously moved the dw and db to dw_cpu and db_cpu in stream_2
                # so we need to synchronize stream_2 before performing Allreduce.
                # In GPU direct, the main stream is already synchronized.

                if not self.model.gpudirect:
                    self.stream_2.synchronize()

                dw_cpu = getattr(self, f"{dw_}_cpu")
                dw_cpu *= self.model.rank_weight
                if self.model.crypt:
                    dw_cpu = self.model.crypt.encrypt(dw_cpu)
                if self.model.use_mpi_buffers:
                    self.model.comm.Allreduce(MPI.IN_PLACE, dw_cpu, op=MPI.SUM)
                else:
                    dw_cpu = self.model.comm.allreduce(dw_cpu, op=MPI.SUM)
                if self.model.crypt:
                    dw_cpu = self.model.crypt.decrypt(dw_cpu)
                setattr(self, f"{dw_}_cpu", dw_cpu)

                if not self.model.gpudirect:
                    dw.ary.set_async(dw_cpu, self.stream_2)

            self.model.tracer.emit_nevent([PYDTNN_MDL_EVENT, PYDTNN_OPS_EVENT], [PYDTNN_EVENT_FINISHED, PYDTNN_EVENT_FINISHED])



from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.multiclass_confusion_matrix import MulticlassConfusionMatrix

import numpy as np

class MulticlassConfusionMatrixCPU(MulticlassConfusionMatrix[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> np.ndarray:
        """
        The output will be a confusion matrix like this:
                |Predicted     |
        ________| 0  | 1  | 2  |
        Target|0| T0 | F1 | F2 |
              |1| F0 | T1 | F2 |
              |2| F0 | F1 | T2 |
        """
        
        # NOTE: y_pred.shape == y_targ.shape == (n<=self.model.batch_size, self.model.output_shape)
        n, target_classes = y_pred.shape
        #assert target_classes == pred_classes, f"target_classes ({target_classes}) != pred_classes {pred_classes}, and must have the same value."
        conf_matrix = np.zeros((target_classes, target_classes), dtype=np.int32)

        for i in range(n):
            target_class = np.nonzero(y_targ[i]== 1)[0]
            predicted_class = np.nonzero(y_pred[i]== 1)[0]
            conf_matrix[target_class, predicted_class] += 1
        
        return conf_matrix



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils.tensor import TensorFormat

if TYPE_CHECKING:
    from pydtnn.model import Model


class Synthetic(Dataset):
    """
    Synthetic Dataset

    Generates random data from:
    - `model.synthetic_train_samples`
    - `model.synthetic_test_samples`
    - `model.synthetic_input_shape` (coma separated)
    - `model.synthetic_output_shape` (coma separated)
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        train_nsamples = int(model.synthetic_train_samples)
        test_nsamples = int(model.synthetic_test_samples)
        input_shape = tuple(map(int, model.synthetic_input_shape.split(",")))
        output_shape = tuple(map(int, model.synthetic_output_shape.split(",")))

        super().__init__(model,
                         train_nsamples=train_nsamples,
                         test_nsamples=test_nsamples,
                         input_shape=input_shape,
                         output_shape=output_shape,
                         force_test_as_validation=force_test_as_validation,
                         debug=debug)

    def _init_actual_data(self):
        self._x = [np.empty((0,)) for part in Dataset.Part]
        self._y = [np.empty((0,)) for part in Dataset.Part]

        for part in Dataset.Part:
            local_batches = self._local_nsamples[part] // self.model.batch_size
            nsamples = local_batches * self.model.batch_size
            x_shape = (nsamples, *self.input_shape)
            x_shape = self.model.encode_shape(x_shape)  # type: ignore
            y_shape = (nsamples, *self.output_shape)
            self._x[part] = np.zeros(x_shape, dtype=self.model.dtype, order="C")
            self._y[part] = np.zeros(y_shape, dtype=self.model.dtype, order="C")

    def _actual_data_generator(self, part: Dataset.Part):
        """
        Generates synthetic data for each dataset part returning (slices of) _x[part] and _y[part] initialized in
        _init_synthetic_data().

        The _local_remaining_nsamples[part] vector is used to keep track of:
        - whether a fresh round of the given part should start (if it is -1), or
        - the remaining number of samples for the given part to be yielded.

        Although the data generator should be called in turns: one round of a part until it finishes, then another
        round of the same or a different part, the current implementation, using -1 to mark the end of a round,
        should also support being called for different parts in an interleaved manner. If another version of this
        method is implemented, at least it should raise and exception if a new round begins when a round for another
        part is still in progress.
        """
        for p in Dataset.Part:
            if self._local_remaining_nsamples[p] == -1:  # If not initialized
                self._local_remaining_nsamples[p] = self._local_nsamples[p]
        while self._local_remaining_nsamples[part] > 0:
            # print()
            # print(f"[part: {part} rank: {self.model.rank}] "
            #       f"{self._local_remaining_nsamples[part]}/{self._x[part].shape[0]}\n")
            if self._local_remaining_nsamples[part] > self._x[part].shape[0]:
                self._local_remaining_nsamples[part] -= self._x[part].shape[0]
                yield self._x[part], self._y[part]
            else:
                remaining_samples = self._local_remaining_nsamples[part]
                self._local_remaining_nsamples[part] = 0
                yield self._x[part][:remaining_samples, ...], self._y[part][:remaining_samples, ...]
        # Mark that a round for part has finished (_local_remaining_nsamples[part] is set to -1 and nothing is yield)
        self._local_remaining_nsamples[part] = -1



from pydtnn.backends.cpu.layers.conv_2d import Conv2DCPU
from pydtnn.backends.cpu.utils.depthwise_conv_nchw_cython import depthwise_conv_backward_nchw_cython, depthwise_conv_nchw_cython
from pydtnn.backends.cpu.utils.depthwise_conv_nhwc_cython import depthwise_conv_backward_nhwc_cython, depthwise_conv_nhwc_cython
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum

import numpy as np

from pydtnn.utils.constants import ArrayShape
from pydtnn.utils.tensor import TensorFormat


class Conv2DDepthwiseCPU(Conv2DCPU):

    def _initializing_special_parameters(self):
        super()._initializing_special_parameters()
        # Setting other parameters
        self.co = self.ci
        # Setting weights
        self.weights_shape = (self.ci, *self.filter_shape)
    # ---

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None):
        super().initialize(prev_shape, x)

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_depthwise_nchw
                self.backward = self._backward_nchw
                _y_shape = (self.model.batch_size, self.co, self.ho, self.wo)
                dx_shape = (self.model.batch_size, self.hi, self.wi, self.ci)
            case TensorFormat.NHWC:
                self.forward = self._forward_depthwise_nhwc
                self.backward = self._backward_nhwc
                _y_shape = (self.model.batch_size, self.ho, self.wo, self.co)
                dx_shape = (self.model.batch_size, self.hi, self.wi, self.ci)
            case _:
                _y_shape = None
                dx_shape = None
                raise NotImplementedError(f"Format \"{self.model.tensor_format}\" is not supported in \"Conv2DDepthwiseCPU\" layer.")
        # ---

        self.dx = np.zeros(shape=dx_shape, dtype=self.model.dtype, order="C")
        self._y = np.zeros(shape=_y_shape, dtype=self.model.dtype, order="C")
    # ---

    def _forward_depthwise_nhwc(self, x: np.ndarray) -> np.ndarray:
        """ Version of the forward that perform a depthwise convolution"""

        self.x = x
        y = self._y[:x.shape[0], ]
        y.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_DEPTHWISE_CONV)
        depthwise_conv_nhwc_cython(x, self.weights, y, self.ho, self.wo,
                                   self.vpadding, self.hpadding,
                                   self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            y: np.ndarray = y.reshape((self.co, -1), copy=False)
            for i in range(self.co):
                np.add(y[i], self.biases[i], out=y[i], 
                       dtype=self.model.dtype)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        y: np.ndarray = y.reshape((-1, self.ho, self.wo, self.co), order="C", copy=None)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)
    
    def _forward_depthwise_nchw(self, x: np.ndarray) -> np.ndarray:
        """ Version of the forward that perform a depthwise convolution"""
        self.x = x
        y = self._y[:x.shape[0], ]
        y.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_DEPTHWISE_CONV)
        depthwise_conv_nchw_cython(x, self.weights, y, self.ho, self.wo,
                                   self.vpadding, self.hpadding,
                                   self.vstride, self.hstride, self.vdilation, self.hdilation)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            y: np.ndarray = y.reshape((self.co, -1), copy=False)
            for i in range(self.co):
                np.add(y[i], self.biases[i], out=y[i], 
                       dtype=self.model.dtype)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_RESHAPE_Y)
        y: np.ndarray = y.reshape((-1, self.co, self.ho, self.wo), order="C", copy=None)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)
    
    def _backward_nhwc(self, dy: np.ndarray) -> np.ndarray:
        
        dx = self.dx[:dy.shape[0], ]
        dx.fill(0)

        depthwise_conv_backward_nhwc_cython(dy, self.x, self.weights,
                                            dx, self.dw,
                                            self.vpadding, self.hpadding,
                                            self.vstride, self.hstride,
                                            self.vdilation, self.hdilation)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 1, 2), out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)
    
    def _backward_nchw(self, dy: np.ndarray) -> np.ndarray:

        dx = self.dx[:dy.shape[0], ]
        dx.fill(0)

        depthwise_conv_backward_nchw_cython(dy, self.x, self.weights,
                                            dx, self.dw,
                                            self.vpadding, self.hpadding,
                                            self.vstride, self.hstride,
                                            self.vdilation, self.hdilation)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 2, 3), out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)
    


from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.abstract.block_layer import AbstractBlockLayer
import numpy as np

class AbstractBlockLayerCPU(AbstractBlockLayer[np.ndarray], LayerCPU):
    pass



"""
PyDTNN Utilities
"""

try:
    from skcuda import cublas  #type: ignore
except Exception:
    pass

import numpy as np


def matmul_gpu(handle, trans_a, trans_b, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc, dtype):
    try:
        gemm = {np.dtype(np.float32): cublas.cublasSgemm,
                np.dtype(np.float64): cublas.cublasDgemm}[dtype]
    except KeyError:
        print("I cannot handle %s type!\n" % dtype.__name__)
    else:
        gemm(handle, trans_a, trans_b, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc)


def matvec_gpu(handle, trans_a, m, n, alpha, a, lda, b, ldb, beta, c, ldc, dtype):
    try:
        gemv = {np.dtype(np.float32): cublas.cublasSgemv,
                np.dtype(np.float64): cublas.cublasDgemv}[dtype]
    except KeyError:
        print("I cannot handle %s type!\n" % dtype.__name__)
    else:
        gemv(handle, trans_a, m, n, alpha, a, lda, b, ldb, beta, c, ldc)



import numpy as np

from pydtnn.activations.tanh import Tanh
from pydtnn.backends.cpu.activations.activation import ActivationCPU


class TanhCPU(Tanh[np.ndarray], ActivationCPU):

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        # NOTE: This attribute only stores data, its value before the operation doesn't matters; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros((self.model.batch_size, *prev_shape), dtype=self.model.dtype, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.y = self._y[:x.shape[0], :]
        np.tanh(x, out=self.y,
                casting="unsafe", dtype=self.model.dtype, order="C")
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        # return 1 - np.tanh(dy) ** 2
        np.tanh(dy, out=dy,
                casting="unsafe", dtype=dy.dtype)
        np.power(dy, 2, out=dy,
                 dtype=self.model.dtype)
        np.subtract(1, dy, out=dy,
                    dtype=self.model.dtype, order="C")

        return dy



import numpy as np

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.conv_2d import Conv2DGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape


from pydtnn.utils.tensor import TensorFormat, format_transpose
from typing import Any, override

from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.memory_allocation import checkConvolutionMemory, getConvolutionWorkspaceSize, getConvolutionWorkspacePtr
import pycuda.gpuarray as gpuarray  #type: ignore

class Conv2DStandardGPU(Conv2DGPU):

    def _initializing_special_parameters(self):
         match self.model.tensor_format:
                case TensorFormat.NCHW:
                    self.weights_shape = (self.co, self.ci, *self.filter_shape)
                case TensorFormat.NHWC:
                    # NOTE: It is this shape, even if in the CPU version is different.
                    self.weights_shape = (self.co, *self.filter_shape, self.ci)
                case _:
                    raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
    # ---

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        # Activations y
        y_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)
        # Derivative dx
        dx_gpu = gpuarray.empty(self.x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Convolution params
        conv_mode = cudnn.cudnnConvolutionMode['CUDNN_CROSS_CORRELATION']
        self.fwd_algo = cudnn.cudnnConvolutionFwdAlgo['CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM']
        self.bwd_dw_algo = cudnn.cudnnConvolutionBwdFilterAlgo['CUDNN_CONVOLUTION_BWD_FILTER_ALGO_1']
        self.bwd_dx_algo = cudnn.cudnnConvolutionBwdDataAlgo['CUDNN_CONVOLUTION_BWD_DATA_ALGO_1']

        # Create convolution descriptor
        self.conv_desc = cudnn.cudnnCreateConvolutionDescriptor()
        cudnn.cudnnSetConvolution2dDescriptor(self.conv_desc, self.vpadding, self.hpadding,
                                              self.vstride, self.hstride, self.vdilation, self.hdilation,
                                              conv_mode, self.model.cudnn_dtype)
        # Set grouping options
        #if self.grouping is Conv2D.Grouping.DEPTHWISE:
        #    cudnn.cudnnSetConvolutionGroupCount(self.conv_desc, self.ci)

        # Allow NCHW -> NHWC conversion for the use of Tensor Cores
        math_type = cudnn.cudnnMathType['CUDNN_TENSOR_OP_MATH_ALLOW_CONVERSION']
        # math_type = cudnn.cudnnMathType['CUDNN_DEFAULT_MATH']
        # math_type = cudnn.cudnnMathType['CUDNN_TENSOR_OP_MATH']
        cudnn.cudnnSetConvolutionMathType(self.conv_desc, math_type)

        # Get output dimensions
        _, _, _ho, _wo = cudnn.cudnnGetConvolution2dForwardOutputDim(self.conv_desc,
                                                                     x.desc, self.weights.desc)
        assert self.ho == _ho and self.wo == _wo, "cuDNN output sizes differ from expected ones!"

        # Set to 20 the number of requested algorithms for enable_cudnn_auto_conv_alg
        req_algs = 20

        self.fwd_algo = cudnn.cudnnFindConvolutionForwardAlgorithm(self.model.cudnn_handle,
                                                                   x.desc, self.weights.desc, self.conv_desc,
                                                                   self.y.desc, req_algs)[0].algo \
            if self.model.enable_cudnn_auto_conv_alg else \
            cudnn.cudnnConvolutionFwdAlgo['CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM']

        local_size = cudnn.cudnnGetConvolutionForwardWorkspaceSize(self.model.cudnn_handle,
                                                                   x.desc, self.weights.desc, self.conv_desc,
                                                                   self.y.desc, self.fwd_algo)
        checkConvolutionMemory(local_size)

        self.bwd_dw_algo = cudnn.cudnnFindConvolutionBackwardFilterAlgorithm(self.model.cudnn_handle,
                                                                             x.desc, self.y.desc, self.conv_desc,
                                                                             self.weights.desc, req_algs)[0].algo \
            if self.model.enable_cudnn_auto_conv_alg else \
            cudnn.cudnnConvolutionBwdFilterAlgo['CUDNN_CONVOLUTION_BWD_FILTER_ALGO_1']

        local_size = cudnn.cudnnGetConvolutionBackwardFilterWorkspaceSize(self.model.cudnn_handle,
                                                                          x.desc, self.y.desc, self.conv_desc,
                                                                          self.weights.desc, self.bwd_dw_algo)
        checkConvolutionMemory(local_size)

        self.bwd_dx_algo = cudnn.cudnnFindConvolutionBackwardDataAlgorithm(self.model.cudnn_handle,
                                                                           self.weights.desc, self.y.desc,
                                                                           self.conv_desc, x.desc,
                                                                           req_algs)[0].algo \
            if self.model.enable_cudnn_auto_conv_alg else \
            cudnn.cudnnConvolutionBwdDataAlgo['CUDNN_CONVOLUTION_BWD_DATA_ALGO_1']

        local_size = cudnn.cudnnGetConvolutionBackwardDataWorkspaceSize(self.model.cudnn_handle,
                                                                        self.weights.desc, self.y.desc,
                                                                        self.conv_desc,
                                                                        x.desc, self.bwd_dx_algo)
        checkConvolutionMemory(local_size)

        self.forward = self._forward_standard
        self.backward = self._backward_standard
    # -----


    def _forward_standard(self, x: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        # Compute a' = x x weights
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        cudnn.cudnnConvolutionForward(self.model.cudnn_handle, alpha,
                                      x.desc, x.ptr,
                                      self.weights.desc, self.weights.ptr,
                                      self.conv_desc, self.fwd_algo,
                                      getConvolutionWorkspacePtr(), getConvolutionWorkspaceSize(), beta,
                                      self.y.desc, self.y.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            alpha, beta = 1.0, 1.0
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT,
                                         self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN_SUM_BIASES)
            # Compute a = a' + biases
            cudnn.cudnnAddTensor(self.model.cudnn_handle, alpha, self.biases.desc, self.biases.ptr,
                                 beta, self.y.desc, self.y.ptr)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y
    # -----

    def _backward_standard(self, dy: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DW)
        # Compute dw
        cudnn.cudnnConvolutionBackwardFilter(self.model.cudnn_handle, alpha,
                                             self.x.desc, self.x.ptr,
                                             dy.desc, dy.ptr, self.conv_desc, self.bwd_dw_algo,
                                             getConvolutionWorkspacePtr(), getConvolutionWorkspaceSize(), beta,
                                             self.dw.desc, self.dw.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        # DtoH dw when data parallelism and no GPU direct/NCCL is used
        if self.model.comm and not self.model.gpudirect and not self.model.enable_nccl:
            self.model.stream.synchronize()
            self.dw.ary.get_async(self.stream_2, self.dw_cpu)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DB)
            # Compute db
            cudnn.cudnnConvolutionBackwardBias(self.model.cudnn_handle, alpha,
                                               dy.desc, dy.ptr, beta,
                                               self.db.desc, self.db.ptr)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

            # DtoH db when data parallelism and no GPU direct/NCCL is used
            if self.model.comm and not self.model.gpudirect and not self.model.enable_nccl:
                self.model.stream.synchronize()
                self.db.ary.get_async(self.stream_2, self.db_cpu)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        # Compute dx
        cudnn.cudnnConvolutionBackwardData(self.model.cudnn_handle, alpha,
                                           self.weights.desc, self.weights.ptr,
                                           dy.desc, dy.ptr,
                                           self.conv_desc, self.bwd_dx_algo,
                                           getConvolutionWorkspacePtr(), getConvolutionWorkspaceSize(), beta,
                                           self.dx.desc, self.dx.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx
    # ----

    @override
    def _export_weights_dw(self, key: str) -> Any:
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: ci, kh, kw, co
                # NCHW's dst: co, ci, kh, kw
                gpu_ary = value.ary
                cpu_ary = gpu_ary.get()
                return np.asarray(format_transpose(cpu_ary, "IHWO", "OIHW"), dtype=np.float64, order="C", copy=True)
            case TensorFormat.NCHW:
                gpu_ary = value.ary
                cpu_ary = gpu_ary.get()
                return cpu_ary
            case default:
                return super()._export_prop(key)
    # ------

    @override
    def _import_weights_dw(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NCHW's src: co, ci, kh, kw
                # NHWC's dst: ci, kh, kw, co
                cpu_ary = np.asarray(format_transpose(value, "OIHW", "IHWO"), dtype=self.model.dtype, order="C", copy=None)
                attribute.ary.set(cpu_ary)
                return
            case default:
                return super()._import_prop(key, value)
    # ---



"""Model test group"""

from warnings import warn

from pydtnn.tests.model_dtype import ModelDTypeTestCase
from pydtnn.tests.model_tensor import ModelTensorTestCase
try:
    from pydtnn.tests.model_gpu import ModelGpuTestCase
except Exception:
    warn("GPU not available, skiping tests!")



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Pooling layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any

# Functionality imports
from pydtnn.layers.max_pool_2d import MaxPool2D
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.adaptive_average_pool_2d import AdaptiveAveragePool2D
import pydtnn.converters.pytorch2pydtnn.common as cm

# ------------------- #
# ---- CONSTANTS ---- #
# ------------------- #

# PyTorch:
PYTORCH_KERNEL_SIZE = "kernel_size"  # INT or Tuple[INT, INT]
PYTORCH_STRIDE = "stride"  # INT or Tuple[INT, INT]
PYTORCH_PADDING = "padding"  # INT or Tuple[INT, INT]
PYTORCH_DILATION = "dilation"  # INT

# PyDTNN:
PYDTNN_POOL_SHAPE = "pool_shape"
PYDTNN_STRIDE = "stride"
PYDTNN_PADDING = "padding"
PYDTNN_DILATION = "dilation"
# ------------------- #

# ------------------- #


def MaxPool2d(args: Dict[str, Any]) -> MaxPool2D:
    # https://pytorch.org/docs/stable/generated/torch.nn.MaxPool2d.html#torch.nn.MaxPool2d

    # PyTorch attributes:
    # Not used: return_indices, ceil_mode
    torch_dict_keys = [PYTORCH_KERNEL_SIZE, PYTORCH_STRIDE, PYTORCH_PADDING, PYTORCH_DILATION]
    # ---- #

    # PyDTNN attributes:
    pydtnn_dict_keys = [PYDTNN_POOL_SHAPE, PYDTNN_STRIDE, PYDTNN_PADDING, PYDTNN_DILATION]
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    if PYDTNN_POOL_SHAPE in layer_args:
        pool_shape = layer_args[PYDTNN_POOL_SHAPE]
        if isinstance(pool_shape, int):
            layer_args[PYDTNN_POOL_SHAPE] = (pool_shape, pool_shape)
        # else: It must be a Tuple[int, int], so it's okay
    # else: Nothing special

    return MaxPool2D(**layer_args)


def AvgPool2d(args: Dict[str, Any]) -> AveragePool2D:
    # https://pytorch.org/docs/stable/generated/torch.nn.AvgPool2d.html#torch.nn.AvgPool2d

    # PyTorch attributes:
    # Not used: ceil_mode, count_include_pad, divisor_override
    torch_dict_keys = [PYTORCH_KERNEL_SIZE, PYTORCH_STRIDE, PYTORCH_PADDING, PYTORCH_DILATION]
    # ---- #

    # PyDTNN attributes:
    pydtnn_dict_keys = [PYDTNN_POOL_SHAPE, PYDTNN_STRIDE, PYDTNN_PADDING, PYDTNN_DILATION]
    # ---- #

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    if PYDTNN_POOL_SHAPE in layer_args:
        pool_shape = layer_args[PYDTNN_POOL_SHAPE]
        if isinstance(pool_shape, int):
            layer_args[PYDTNN_POOL_SHAPE] = (pool_shape, pool_shape)
        # else: It must be a Tuple[int, int], so it's okay

    return AveragePool2D(**layer_args)


def AdaptiveAvgPool2d(args: Dict[str, Any]) -> AdaptiveAveragePool2D:
    # https://pytorch.org/docs/stable/generated/torch.nn.AdaptiveAvgPool2d.html#torch.nn.AdaptiveAvgPool2d
    # from torch.nn import AdaptiveAvgPool2d

    arguments = args[cm.ARGUMENTS]
    output_shape = arguments[cm.PYTORCH_OUTPUT_SIZE] if cm.PYTORCH_OUTPUT_SIZE in arguments else None

    return AdaptiveAveragePool2D(output_shape=output_shape)



from abc import abstractmethod

from pydtnn.backends import PromoteToBackend
from pydtnn.utils import find_component
from pydtnn.utils.constants import ArrayShape, Array
import numpy as np

class Metric[T: Array](PromoteToBackend):
    format = ""
    order = 0   # No need of special order.

    def __init__(self, shape: ArrayShape, eps=1e-8):
        self.shape = shape
        self.eps = eps

    def initialize(self) -> None:
        pass

    @abstractmethod
    def compute(self, y_pred: T, y_targ: T) -> float | np.ndarray:
        pass


def select(name: str) -> type[Metric]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



import inspect

import numpy as np

from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython
from pydtnn.tests.abstract.common import verbose_test, D, alexnet_layers, TestCase
from pydtnn.utils import print_with_header, random


class ConvCommonTestCase(TestCase):
    """
    Tests that conv leads to the same results as i2c and mm.
    """

    @classmethod
    def _compute_both(cls, weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None,
                      vpadding=0, hpadding=0, vstride=1, hstride=1,
                      vdilation=1, hdilation=1) -> tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()

    @staticmethod
    def _get_config() -> D:
        return D()

    @staticmethod
    def _compute(weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None, kh=1, kw=1, vpadding=0, hpadding=0, vstride=1, hstride=1, vdilation=1, hdilation=1):
        raise NotImplementedError()

    def test_raise_on_different_strides(self):
        d = self._get_config()
        weights = np.ones((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = np.ones((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        test_result, ref_result = self._compute_both(weights, x,
                                                     vpadding=d.vpadding, hpadding=d.hpadding,
                                                     vstride=1, hstride=2,
                                                     vdilation=d.vdilation, hdilation=d.hdilation)
        self.assertTrue(np.allclose(test_result, ref_result))

    def test_defaults_with_ones(self):
        """
        Test that the default parameters on ones matrices lead to the same solution
        """
        d = self._get_config()
        weights = np.ones((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = np.ones((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        test_result, ref_result = self._compute_both(weights, x,
                                                     vpadding=d.vpadding, hpadding=d.hpadding,
                                                     vstride=d.vstride, hstride=d.hstride,
                                                     vdilation=d.vdilation, hdilation=d.hdilation)
        self.assertTrue(np.allclose(test_result, ref_result))

    def test_defaults_with_random(self):
        """
        Test that the default parameters on random matrices lead to the same solution
        """
        d = self._get_config()
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        test_result, ref_result = self._compute_both(weights, x,
                                                     vpadding=d.vpadding, hpadding=d.hpadding,
                                                     vstride=d.vstride, hstride=d.hstride,
                                                     vdilation=d.vdilation, hdilation=d.hdilation)
        self.assertTrue(np.allclose(test_result, ref_result))

    def test_defaults_including_biases_with_random(self):
        """
        Test that the default parameters on random matrices, including b, lead to the same solution
        """
        d = self._get_config()
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        ho = (d.h + 2 * d.vpadding - d.vdilation * (d.kh - 1) - 1) // d.vstride + 1
        wo = (d.w + 2 * d.hpadding - d.hdilation * (d.kw - 1) - 1) // d.hstride + 1
        #biases = random.random((d.b, ho, wo, d.kn)).astype(np.float32, order='C')
        biases = random.random((d.kn, )).astype(np.float32, order='C')
        test_result, ref_result = self._compute_both(weights, x, biases=biases,
                                                     vpadding=d.vpadding, hpadding=d.hpadding,
                                                     vstride=d.vstride, hstride=d.hstride,
                                                     vdilation=d.vdilation, hdilation=d.hdilation)
        diff = test_result - ref_result
        self.assertTrue(np.allclose(test_result, ref_result), f"The difference is to big (rtol=1.e-5, atol=1.e-8). max diff: {diff.max()}. min diff: {diff.min()}")

    def test_with_different_kn(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print(" kn   Maximum difference    sum(cg_result)")
            print("----+--------------------+-----------------")
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        np_all_close_for_all_cases = True
        for kn in range(1, 32):
            weights = random.random((d.c, d.kh, d.kw, kn)).astype(np.float32, order='C')
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=d.kw, kh=d.kh,
                                                    vpadding=d.vpadding, hpadding=d.hpadding,
                                                    vstride=d.vstride, hstride=d.hstride,
                                                    vdilation=d.vdilation, hdilation=d.hdilation)
            test_result: np.ndarray = test_result.reshape((-1, kn), copy=False)
            n, h, w, c = x.shape

            ho = (h + 2 * d.vpadding - d.vdilation * (d.kh - 1) - 1) // d.vstride + 1
            wo = (w + 2 * d.hpadding - d.hdilation * (d.kw - 1) - 1) // d.hstride + 1

            dim_n = n * ho * wo
            dim_c = c * d.kh * d.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               d.kh, d.kw, ho, wo,
                               d.vpadding, d.hpadding,
                               d.vstride, d.hstride,
                               d.vdilation, d.hdilation)
            w_c = weights.reshape((-1, kn), copy=False)
            ref_result = x_c @ w_c
            if verbose_test():
                print("{:3}    {:9.7f}             {:11.2f}"
                      "".format(kn, max([abs(x - y) for x, y in zip(test_result.flatten(),
                                                                    ref_result.flatten())]),
                                np.sum(test_result)))
            np_all_close_for_all_cases = np_all_close_for_all_cases and np.allclose(test_result,
                                                                                    ref_result)
        self.assertTrue(np_all_close_for_all_cases)

    def test_with_different_b(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print("  b   Maximum difference")
            print("----+--------------------")
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        np_all_close_for_all_cases = True
        for b in range(1, 32):
            x = random.random((b, d.h, d.w, d.c)).astype(np.float32, order='C')
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=d.kw, kh=d.kh,
                                                    vpadding=d.vpadding, hpadding=d.hpadding,
                                                    vstride=d.vstride, hstride=d.hstride,
                                                    vdilation=d.vdilation, hdilation=d.hdilation)
            test_result: np.ndarray = test_result.reshape((-1, d.kn), copy=False)
            n, h, w, c = x.shape

            ho = (h + 2 * d.vpadding - d.vdilation * (d.kh - 1) - 1) // d.vstride + 1
            wo = (w + 2 * d.hpadding - d.hdilation * (d.kw - 1) - 1) // d.hstride + 1

            dim_n = n * ho * wo
            dim_c = c * d.kh * d.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               d.kh, d.kw, ho, wo,
                               d.vpadding, d.hpadding,
                               d.vstride, d.hstride,
                               d.vdilation, d.hdilation)
            w_c = weights.reshape((-1, d.kn), copy=False)
            ref_result = x_c @ w_c
            if verbose_test():
                print("{:3}    {:9.7f}".format(b,
                                               max([abs(x - y) for x, y
                                                    in
                                                    zip(test_result.flatten(), ref_result.flatten())])))
            np_all_close_for_all_cases = np_all_close_for_all_cases and np.allclose(test_result,
                                                                                    ref_result)
        self.assertTrue(np_all_close_for_all_cases)

    def test_with_different_padding(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print("  p   Maximum difference")
            print("----+--------------------")
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        np_all_close_for_all_cases = True
        for padding in range(0, 5):
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=d.kw, kh=d.kh,
                                                    vpadding=padding, hpadding=padding,
                                                    vstride=d.vstride, hstride=d.hstride,
                                                    vdilation=d.vdilation, hdilation=d.hdilation)
            test_result: np.ndarray = test_result.reshape((-1, d.kn), copy=False)
            n, h, w, c = x.shape

            ho = (h + 2 * padding - d.vdilation * (d.kh - 1) - 1) // d.vstride + 1
            wo = (w + 2 * padding - d.hdilation * (d.kw - 1) - 1) // d.hstride + 1

            dim_n = n * ho * wo
            dim_c = c * d.kh * d.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               d.kh, d.kw, ho, wo,
                               padding, padding,
                               d.vstride, d.hstride,
                               d.vdilation, d.hdilation)
            w_c = weights.reshape((-1, d.kn), copy=False)
            ref_result = x_c @ w_c
            if verbose_test():
                print("{:3}    {:9.7f}".format(padding,
                                               max([abs(x - y) for x, y
                                                    in
                                                    zip(test_result.flatten(), ref_result.flatten())])))
            np_all_close_for_all_cases = np_all_close_for_all_cases and np.allclose(test_result,
                                                                                    ref_result)
        self.assertTrue(np_all_close_for_all_cases)

    def test_with_different_stride(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print("  s   Maximum difference")
            print("----+--------------------")
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        np_all_close_for_all_cases = True
        for stride in range(1, 6):
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=d.kw, kh=d.kh,
                                                    vpadding=d.vpadding, hpadding=d.hpadding,
                                                    vstride=stride, hstride=stride,
                                                    vdilation=d.vdilation, hdilation=d.hdilation)
            test_result: np.ndarray = test_result.reshape((-1, d.kn), copy=False)
            n, h, w, c = x.shape

            ho = (h + 2 * d.vpadding - d.vdilation * (d.kh - 1) - 1) // stride + 1
            wo = (w + 2 * d.hpadding - d.hdilation * (d.kw - 1) - 1) // stride + 1

            dim_n = n * ho * wo
            dim_c = c * d.kh * d.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               d.kh, d.kw, ho, wo,
                               d.vpadding, d.hpadding,
                               stride, stride,
                               d.vdilation, d.hdilation)
            w_c = weights.reshape((-1, d.kn), copy=False)
            ref_result = x_c @ w_c
            if verbose_test():
                print("{:3}    {:9.7f}".format(stride,
                                               max([abs(x - y) for x, y
                                                    in
                                                    zip(test_result.flatten(), ref_result.flatten())])))
            np_all_close_for_all_cases = np_all_close_for_all_cases and np.allclose(test_result,
                                                                                    ref_result)
        self.assertTrue(np_all_close_for_all_cases)

    def test_with_different_strides(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print(" vs  hs   Maximum difference")
            print("--------+--------------------")
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        for vstride in range(1, 5):
            for hstride in range(1, 5):
                if vstride == hstride:
                    continue
                test_result: np.ndarray = self._compute(weights, x,
                                                        kw=d.kw, kh=d.kh,
                                                        vpadding=d.vpadding, hpadding=d.hpadding,
                                                        vstride=vstride, hstride=hstride,
                                                        vdilation=d.vdilation, hdilation=d.hdilation)
                test_result: np.ndarray = test_result.reshape((-1, d.kn), copy=False)
                n, h, w, c = x.shape

                ho = (h + 2 * d.vpadding - d.vdilation * (d.kh - 1) - 1) // vstride + 1
                wo = (w + 2 * d.hpadding - d.hdilation * (d.kw - 1) - 1) // hstride + 1

                dim_n = n * ho * wo
                dim_c = c * d.kh * d.kw

                x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

                # FIXME: To abstract method (or use compute both)
                im2row_nhwc_cython(x, x_c,
                                   d.kh, d.kw, ho, wo,
                                   d.vpadding, d.hpadding,
                                   vstride, hstride,
                                   d.vdilation, d.hdilation)
                w_c = weights.reshape((-1, d.kn), copy=False)
                ref_result = x_c @ w_c
                if verbose_test():
                    print("{:3} {:3}    {:9.7f}".format(vstride, hstride,
                                                        max([abs(x - y) for x, y
                                                             in
                                                             zip(test_result.flatten(),
                                                                 ref_result.flatten())])))
                self.assertTrue(np.allclose(test_result, ref_result),
                                f"Results differ with vstride {vstride} and hstride {hstride}")

    def test_with_different_dilation(self):
        d = self._get_config()
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print("  s   Maximum difference")
            print("----+--------------------")
        weights = random.random((d.c, d.kh, d.kw, d.kn)).astype(np.float32, order='C')
        x = random.random((d.b, d.h, d.w, d.c)).astype(np.float32, order='C')
        np_all_close_for_all_cases = True
        for dilation in range(1, 3):
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=d.kw, kh=d.kh,
                                                    vpadding=d.vpadding, hpadding=d.hpadding,
                                                    vstride=d.vstride, hstride=d.hstride,
                                                    vdilation=dilation, hdilation=dilation)
            test_result: np.ndarray = test_result.reshape((-1, d.kn), copy=False)
            n, h, w, c = x.shape

            ho = (h + 2 * d.vpadding - dilation * (d.kh - 1) - 1) // d.vstride + 1
            wo = (w + 2 * d.hpadding - dilation * (d.kw - 1) - 1) // d.hstride + 1

            dim_n = n * ho * wo
            dim_c = c * d.kh * d.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               d.kh, d.kw, ho, wo,
                               d.vpadding, d.hpadding,
                               d.vstride, d.hstride,
                               dilation, dilation)
            w_c = weights.reshape((-1, d.kn), copy=False)
            ref_result = x_c @ w_c
            if verbose_test():
                print("{:3}    {:9.7f}".format(dilation,
                                               max([abs(x - y) for x, y
                                                    in
                                                    zip(test_result.flatten(), ref_result.flatten())])))
            np_all_close_for_all_cases = np_all_close_for_all_cases and np.allclose(test_result,
                                                                                    ref_result)
        self.assertTrue(np_all_close_for_all_cases)

    def test_alexnet_layers(self):
        if verbose_test():
            print_with_header("{}".format(inspect.stack()[1][3]), None)
            print(" layer   Maximum difference")
            print("-------+--------------------")
        layers = alexnet_layers
        for n, layer in enumerate(layers):
            weights = random.random((layer.c, layer.kh, layer.kw, layer.kn)).astype(np.float32, order='C')
            x = random.random((layer.b, layer.h, layer.w, layer.c)).astype(np.float32, order='C')
            test_result: np.ndarray = self._compute(weights, x,
                                                    kw=layer.kw, kh=layer.kh,
                                                    vpadding=layer.vpadding, hpadding=layer.hpadding,
                                                    vstride=layer.vstride, hstride=layer.hstride,
                                                    vdilation=layer.vdilation, hdilation=layer.hdilation)
            test_result: np.ndarray = test_result.reshape(-1, layer.kn)
            n, h, w, c = x.shape

            dim_n = n * layer.ho * layer.wo
            dim_c = c * layer.kh * layer.kw

            x_c = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

            # FIXME: To abstract method (or use compute both)
            im2row_nhwc_cython(x, x_c,
                               layer.kh, layer.kw, layer.ho, layer.wo,
                               layer.vpadding, layer.hpadding,
                               layer.vstride, layer.hstride,
                               layer.vdilation, layer.hdilation)
            w_c = weights.reshape(-1, layer.kn)
            ref_result = x_c @ w_c
            if verbose_test():
                print("   {:2}      {:9.7f}".format(n,
                                                    max([abs(x - y) for x, y
                                                         in
                                                         zip(test_result.flatten(),
                                                             ref_result.flatten())])))
                if n == 9:
                    print("Flags for last test_result output:")
                    print(test_result.flags)
            self.assertTrue(np.allclose(test_result, ref_result),
                            f"Results differ for AlexNet Cifar and ImageNet layers number {n}")



import numpy as np

from pydtnn.backends.cpu.layers.conv_2d import Conv2DCPU
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_OPS_EVENT_enum, PYDTNN_EVENT_FINISHED
from pydtnn.utils.best_transpose_0231 import best_transpose_0231
from pydtnn.utils.best_transpose_0312 import best_transpose_0312
from pydtnn.model import Model
from pydtnn.utils.constants import ArrayShape, Parameters
from pydtnn.utils.tensor import TensorFormat, format_transpose


class Conv2DPointwiseCPU(Conv2DCPU):

    def _export_prop(self, key: str):
        if key not in {Parameters.WEIGHTS, Parameters.DW}:
            return super()._export_prop(key)
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: ci, co
                # NCHW's dst: co, ci
                return np.asarray(format_transpose(value, "IO", "OI"), dtype=np.float64, order="C", copy=True)
        return super()._export_prop(key)
    # -----

    def _import_prop(self, key: str, value) -> None:
        if key not in {Parameters.WEIGHTS, Parameters.DW}:
            return super()._import_prop(key, value)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NCHW's src: co, ci
                # NHWC's dst: ci, co
                ary = getattr(self, key)
                ary[:] = np.asarray(format_transpose(value, "OI", "IO"), dtype=self.model.dtype, order="C", copy=None)
                return
        return super()._import_prop(key, value)
    # ------

    def _initializing_special_parameters(self):
        super()._initializing_special_parameters()
        # Setting other parameters
        self.kh = self.kw = 1
        # Setting weights
        match self.model.tensor_format:
                case TensorFormat.NCHW:
                    self.weights_shape = (self.co, self.ci)
                case TensorFormat.NHWC:
                    self.weights_shape = (self.co, self.ci)
                case _:
                    raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
        #--
    # ---

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None) -> None:
        super().initialize(prev_shape, x)
        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_pointwise_nchw
                self.backward = self._backward_nchw
            case TensorFormat.NHWC:
                self.forward = self._forward_pointwise_nhwc
                self.backward = self._backward_nhwc
        # --
        y_shape = self.model.encode_shape((self.model.batch_size, self.co, self.ho, self.wo))
        # NOTE: These attributes only store data, their values before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        # self.dw (this one too, but it's initalized in Conv2DCPU)
        self.y = np.zeros(shape=y_shape, dtype=self.model.dtype, order="C")
        self.dx = np.zeros(shape=(self.ci, self.model.batch_size * self.hi * self.wi), dtype=self.model.dtype, order="C")
    # ------

    def _forward_pointwise_nhwc(self, x: np.ndarray) -> np.ndarray:
        if self.model.mode is Model.Mode.TRAIN:
            self.x: np.ndarray = x

        y = self.y[:x.shape[0], :]
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_POINTWISE_CONV)
        np.matmul(x, self.weights, out=y,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            np.add(y, self.biases.reshape((1, 1, 1, self.co), copy=False), out=y,
                   dtype=self.model.dtype, order="C")
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _forward_pointwise_nchw(self, x: np.ndarray) -> np.ndarray:

        if self.model.mode is Model.Mode.TRAIN:
            self.x: np.ndarray = x

        y = self.y[:x.shape[0], :]

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_TRANSPOSE_Y)
        y = best_transpose_0231(y)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_POINTWISE_CONV)
        np.matmul(best_transpose_0231(x), self.weights.T, out=y,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_TRANSPOSE_Y)
        y: np.ndarray = best_transpose_0312(y)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_SUM_BIASES)
            np.add(y, self.biases.reshape((1, self.co, 1, 1), copy=False), out=y,
                   dtype=self.model.dtype, order="C")
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward_nhwc(self, dy: np.ndarray) -> np.ndarray:

        _n, _h, _w, _c = dy.shape
        _dim = _n * _h * _w
        x_shape = self.x.shape
        dx = np.asarray(self.dx[:, :_dim], dtype=self.model.dtype, order="C", copy=None)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_DY)
        reshaped_dy = dy.reshape((_dim, _c), copy=False)
        self.x = self.x.reshape((-1, _dim), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DW_MATMUL)
        np.matmul(self.x, reshaped_dy, out=self.dw,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 1, 2), out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_W)
        w = self.weights.reshape((self.co, -1), copy=False).T
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        reshaped_dy: np.ndarray = dy.reshape((self.co, -1), copy=False)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_MATMUL)
        np.matmul(w, reshaped_dy, out=dx,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx.reshape(x_shape, copy=False), dtype=self.model.dtype, order='C', copy=None)

    def _backward_nchw(self, dy: np.ndarray) -> np.ndarray:

        _n, _c, _h, _w = dy.shape
        _dim = _n * _h * _w
        x_shape = self.x.shape
        dx = np.asarray(self.dx[:, :_dim], dtype=self.model.dtype, order="C", copy=None)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_DY)
        reshaped_dy = dy.reshape((_dim, _c), copy=False)
        self.x = self.x.reshape((-1, _dim), copy=False)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DW_MATMUL)
        np.matmul(self.x, reshaped_dy, out=self.dw.T,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_SUM_BIASES)
            np.sum(dy, axis=(0, 2, 3), out=self.db)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_TRANSPOSE_W)
        w = self.weights.reshape((self.co, -1), copy=False).T
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        reshaped_dy: np.ndarray = dy.reshape((self.co, -1), copy=False)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.COMP_DX_MATMUL)
        np.matmul(w, reshaped_dy, out=dx,
                  dtype=self.model.dtype, order="C")
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(dx.reshape(x_shape, copy=False), dtype=self.model.dtype, order='C', copy=None)



import numpy as np

from pydtnn.activations.arctanh import Arctanh
from pydtnn.backends.cpu.activations.activation import ActivationCPU


class ArctanhCPU(Arctanh[np.ndarray], ActivationCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self, prev_shape, x=None):
        super().initialize(prev_shape, x)
        # NOTE: This attribute only stores data, its value before the operation doesn't matters; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self._y = np.zeros(shape=(self.model.batch_size, *self.shape),
                           dtype=self.model.dtype, order="C")

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.y = self._y[:x.shape[0], :]
        np.arctanh(x, out=self.y, casting="unsafe", dtype=self.model.dtype, order="C")
        return self.y

    def backward(self, dy: np.ndarray) -> np.ndarray:
        # return 1 / (1 + dy ** 2)
        np.power(dy, 2, out=dy,
                 casting="unsafe", dtype=self.model.dtype)
        np.add(dy, 1, out=dy,
               casting="unsafe", dtype=self.model.dtype)
        np.reciprocal(dy, out=dy, casting="unsafe", dtype=self.model.dtype, order="C")
        return dy



import numpy as np
from pydtnn.backends.cpu.utils.bn_training_cython import bn_training_bwd_cython  # , bn_training_fwd_cython

from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.model import Model
from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.utils.tensor import TensorFormat, format_transpose
from pydtnn.utils.constants import ArrayShape, Parameters


class BatchNormalizationCPU(BatchNormalization[np.ndarray], LayerCPU):

    @property
    def _ary_prop(self) -> set[str]:
        return {Parameters.RUNNING_MEAN,
                Parameters.RUNNING_VAR,
                *super()._ary_prop}

    @staticmethod
    def get_inv_std(running_var: np.ndarray, epsilon: float, dtype: np.dtype) -> np.ndarray:
        inv_std = np.add(running_var, epsilon, dtype=dtype, order="C")
        np.sqrt(inv_std, out=inv_std,
                dtype=dtype)
        np.reciprocal(inv_std, out=inv_std,
                      dtype=dtype)
        return inv_std

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)

        if self.spatial:
            self.ci, self.hi, self.wi = self.model.decode_shape(self.shape)
            shape_ = (self.ci,)
        else:
            self.ci = self.shape[0]
            shape_ = (self.ci,)

        self.gamma = np.full(shape_, self.gamma_init_val, dtype=self.model.dtype, order="C")
        self.beta = np.full(shape_, self.beta_init_val, dtype=self.model.dtype, order="C")
        self.dgamma: np.ndarray = np.zeros(shape=(self.ci,), dtype=self.model.dtype, order="C")
        self.dbeta: np.ndarray = np.zeros(shape=(self.ci,), dtype=self.model.dtype, order="C")
        self.running_mean = self.moving_mean_initializer(shape_, self.model.dtype)
        self.running_var = self.moving_variance_initializer(shape_, self.model.dtype)
        self.nparams = self.gamma.size + self.beta.size + self.running_mean.size + self.running_var.size

        # NOTE: These attributes only store data, their value before the operation doesn't matter; they're initalized due avoid warnings in "LayerAndActivationBase.export".
        self._mean_inv: np.ndarray = np.zeros(shape=(self.ci,), dtype=self.model.dtype, order="C")
        self._var_inv: np.ndarray = np.zeros(shape=(self.ci,), dtype=self.model.dtype, order="C")
        self.std: np.ndarray = np.zeros(shape=(self.ci,), dtype=self.model.dtype, order="C")

        if self.spatial:
            self.dx: np.ndarray = np.zeros(shape=(self.model.batch_size * self.hi * self.wi, self.ci), dtype=self.model.dtype, order="C")
            self.y = np.zeros((self.model.batch_size * self.hi * self.wi, self.ci), dtype=self.model.dtype, order="C")
            self.dy_xn = np.zeros((self.model.batch_size * self.hi * self.wi, self.ci), dtype=self.model.dtype, order="C")
        else:
            # NOTE: in this case, self.hi and self.wi are 0 (self.shape should be somethin like: "(512, )"
            self.dx: np.ndarray = np.zeros(shape=(self.model.batch_size, self.ci), dtype=self.model.dtype, order="C")
            self.y = np.zeros((self.model.batch_size, self.ci), dtype=self.model.dtype, order="C")
            self.dy_xn = np.zeros((self.model.batch_size, self.ci), dtype=self.model.dtype, order="C")
    # --

    def forward(self, x: np.ndarray) -> np.ndarray:

        self.y: np.ndarray
        n = x.shape[0]

        if self.spatial:
            # NOTE: Executing in this format gives better results.
            x = format_transpose(x, self.model.tensor_format, TensorFormat.NHWC)
            x = x.reshape((-1, self.ci), copy=None, order="C")
        # else: x = x (no reshape needed)

        y: np.ndarray = self.y[:x.shape[0], :]
        self.xn = x

        if self.model.mode is Model.Mode.EVALUATE:
            _mean = self.running_mean
            _var = self.running_var
        else:  # ModelModeEnum.TRAIN:
            _mean = np.mean(self.xn, axis=0, dtype=self.model.dtype)
            _var = np.var(self.xn, axis=0, dtype=self.model.dtype)

            inv_momentum = (1.0 - self.momentum)
            # self.running_mean = self.momentum * self.running_mean + inv_momentum * _mean
            np.multiply(self.momentum, self.running_mean, out=self.running_mean,
                        dtype=self.model.dtype)
            np.multiply(inv_momentum, _mean, out=self._mean_inv,
                        dtype=self.model.dtype, order="C")
            np.add(self.running_mean, self._mean_inv, out=self.running_mean,
                   dtype=self.model.dtype)

            # self.running_var = self.momentum * self.running_var + inv_momentum * _var
            np.multiply(self.momentum, self.running_var, out=self.running_var,
                        dtype=self.model.dtype)
            np.multiply(inv_momentum, _var, out=self._var_inv,
                        dtype=self.model.dtype, order="C")
            np.add(self.running_var, self._var_inv, out=self.running_var,
                   dtype=self.model.dtype)
        # anyways:

        # bn_training_fwd_cython(x, y, self.xn, self.std, self.gamma, self.beta, _mean, _var, self.epsilon)
        # y = ((x - mean(x)) / sqrt(var(x) + epsilon)) * gamma + beta
        np.subtract(self.xn, _mean, out=self.xn,
                    dtype=self.model.dtype)

        np.add(_var, self.epsilon, out=self.std,
               dtype=self.model.dtype, order="C")
        np.sqrt(self.std, out=self.std,
                dtype=self.model.dtype)

        np.divide(self.xn, self.std, out=self.xn,
                  dtype=self.model.dtype)
        np.multiply(self.gamma, self.xn, out=y,
                    dtype=self.model.dtype)
        np.add(y, self.beta, out=y,
               dtype=self.model.dtype)

        if self.spatial:
            y = y.reshape((n, self.hi, self.wi, self.ci), copy=False)
            y = format_transpose(y, TensorFormat.NHWC, self.model.tensor_format)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def backward(self, dy: np.ndarray) -> np.ndarray:

        n = dy.shape[0]
        if self.spatial:
            num_elems = (n * self.hi * self.wi)

            # NOTE: Executing in this format gives better results.
            dy = format_transpose(dy, self.model.tensor_format, TensorFormat.NHWC)
            dy = dy.reshape((num_elems, self.ci), copy=None)
        else:
            num_elems = n

        dx: np.ndarray = self.dx[: num_elems, :]
        dy_xn: np.ndarray = self.dy_xn[: num_elems, :]

        # dx = (self.gamma / (self.std * n)) * (n * dy - self.xn * self.dgamma - self.dbeta)
        np.multiply(dy, self.xn, out=dy_xn, dtype=self.model.dtype)
        np.sum(dy_xn, axis=0, out=self.dgamma, dtype=self.model.dtype)
        np.sum(dy, axis=0, out=self.dbeta, dtype=self.model.dtype)

        bn_training_bwd_cython(dx, dy, self.xn, self.std, self.gamma, self.dgamma, self.dbeta)

        if self.spatial:
            dx = dx.reshape((n, self.hi, self.wi, self.ci), copy=False)
            dx = format_transpose(dx, TensorFormat.NHWC, self.model.tensor_format)

        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)



import enum
import importlib
import typing

if typing.TYPE_CHECKING:
    from pydtnn import model as model_module


class BackendType(enum.StrEnum):
    CPU = enum.auto()
    GPU = enum.auto()


class PromoteToBackend:
    _backend: typing.Self

    def __new__(cls, *args, **kwds):
        # Save top-level constructor arguments
        self = super().__new__(cls)
        self._new_backend = (args, kwds)  # type: ignore
        return self

    def __getattribute__(self, name: str):
        ref = "_backend"

        # Get backend
        try:
            backend = super().__getattribute__(ref)
        except AttributeError:
            backend = None

        # Skip backend if internal
        if backend is None or name.endswith(ref):
            return super().__getattribute__(name)
        else:
            return getattr(backend, name)

    def _get_backend_cls(self) -> typing.Any:
        cls = self.__class__
        backend = self.model._backend
        module_name = cls.__module__.split(".", 1)[1]
        backend_module_name = f"pydtnn.backends.{backend}.{module_name}"
        backend_module = importlib.import_module(backend_module_name)
        cls_name = f"{cls.__name__}{backend.upper()}"
        cls = getattr(backend_module, cls_name)
        return cls

    def __setattr__(self, name: str, value) -> None:
        ref = "_backend"

        # Get backend
        backend = getattr(self, ref, None)

        # Skip backend if internal
        if backend is None or name.endswith(ref):
            super().__setattr__(name, value)
        else:
            setattr(backend, name, value)

    def __delattr__(self, name: str) -> None:
        ref = "_backend"

        # Get backend
        backend = getattr(self, ref, None)

        # Skip backend if internal
        if backend is None or name.endswith(ref):
            super().__delattr__(name)
        else:
            delattr(backend, name)

    def init_backend(self) -> None:
        """
        Initialize the backend implementation used

        **Notice**: All object attributes are cleared when called.
        So, if used, this method should be the first called.
        """
        # Clear backend
        try:
            del self._backend
        except AttributeError:
            pass

        # Get backend class
        cls = self._get_backend_cls()
        if cls is None:
            return

        # Create backend instance
        args, kwds = self._new_backend
        self._backend = cls(*args, **kwds)
        self._frontend = self

    # Base class
    model: "model_module.Model"

    def set_model(self, model: "model_module.Model") -> None:
        """Link a to a new model instance"""
        self.model = model

    def init_backend_from_model(self, model: "model_module.Model") -> None:
        """Initialize backend and link a new model instance"""
        #  NOTE: This one is to have access to "model" in the "get_backend_cls"
        self.set_model(model)
        #  NOTE: This function masks all previous attribute data.
        self.init_backend()
        #  NOTE: This one is to set model in the backend class.
        self.set_model(model)
    # ---

    @classmethod
    def from_model[C](cls: type[C], model: "model_module.Model") -> C:
        raise NotImplementedError("Use a concrete optimizer!")



import pycuda.gpuarray as gpuarray  # type: ignore
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore

from pydtnn.losses.binary_cross_entropy import BinaryCrossEntropy
from pydtnn.backends.gpu.losses.loss import LossGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE


class BinaryCrossEntropyGPU(LossGPU, BinaryCrossEntropy[TensorGPU]):

    def __init_gpu_kernel__(self) -> Function:
        _name = "binary_cross_entropy"
        code = """
        __global__ void {name}({T} *y_targ, {T} *y_pred, {T} *res,
                               {T} *dx, int b, int n, {T} eps)
        {{
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < b) {
                int i = 0, max = 0;
                {T} pred;
                res[idx] = 0;
                for ( i = 0; i < n; i++ ) 
                {{
                    res[idx]+= logf(fmaxf((1 - y_targ[idx * n + i] ) -
                                               y_pred[idx * n + i], eps));
                    pred = y_pred[idx * n + max];
                    if ( pred < eps )          pred = eps;
                    else if ( pred > (1-eps) ) pred = (1-eps);
                    dx[idx * n + i] = (-(y_targ[idx * n + i]  / pred) +
                                   ((1 - y_targ[idx * n + i]) / pred) ) / b;
                }}
            }}
            return;
        }}
        """.format(
            T = DTYPE2CTYPE[self.model.dtype],
            name = _name
            )
        
        module = SourceModule(code).get_function(_name)
        return module

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU, batch_size: int) -> tuple[float, TensorGPU]:
        assert len(y_targ.shape) == 2
        self.kernel(y_targ, y_pred, self.loss, self.dx.ary,
                    batch_size, self.shape[1], self.eps,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        loss: float = -gpuarray.sum(self.loss[:batch_size]) / batch_size
        return loss, self.dx



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.categorical_hinge import CategoricalHinge


class CategoricalHingeCPU(CategoricalHinge[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        # pos = np.sum(y_targ * y_pred, axis=-1)
        # neg = np.max((1.0 - y_targ) * y_pred, axis=-1)
        # return np.mean(np.maximum(0.0, neg - pos + 1), axis=-1)

        pos:np.ndarray = np.multiply(y_targ, y_pred, dtype=self.model.dtype)
        pos = np.sum(pos, axis=-1, dtype=self.model.dtype)

        neg = np.multiply(-1, y_targ, dtype=self.model.dtype)
        np.add(neg, 1, out=neg, dtype=self.model.dtype)
        np.multiply(neg, y_pred, out= neg, dtype=self.model.dtype)
        neg:np.ndarray = np.max(neg, axis=-1)

        np.subtract(neg, pos, out=neg, dtype=self.model.dtype)
        np.add(neg, 1, out=neg, dtype=self.model.dtype)
        maximum = np.maximum(0.0, neg)

        maximum = np.mean(maximum, axis=-1)

        return maximum
